from bson.int64 import Int64

from mindsdb_sql.parser.ast import Identifier, Insert, CreateTable

from mindsdb.api.mongo.classes import Responder
import mindsdb.api.mongo.functions as helpers
from mindsdb.api.mongo.responders.find import find_to_ast

from mindsdb.api.mongo.classes.query_sql import run_sql_command


def aggregate_to_ast(query, database):
    collection = query['aggregate']

    save_table = None
    is_append = False
    first_step = query['pipeline'][0]
    if '$match' in first_step:
        # convert to find
        find_query = {
            'find': collection,
            'filter': first_step['$match'],
        }
        for step in query['pipeline'][1:]:
            if '$project' in step:
                find_query['projection'] = step['$project']
            if '$sort' in step:
                find_query['sort'] = step['$sort']
            if '$skip' in step:
                find_query['skip'] = step['$skip']
            if '$limit' in step:
                find_query['limit'] = step['$limit']

            if '$out' in step:
                target = step['$out']
                if isinstance(target, str):
                    save_table = Identifier(target)
                else:
                    save_table = Identifier(parts=[target['db'], target['coll']])
                    if target.get('append'):
                        is_append = True

            # TODO implement group
        ast_query = find_to_ast(find_query, database)
        if save_table is not None:
            if is_append:
                ast_query = Insert(save_table, from_select=ast_query)
            else:
                ast_query = CreateTable(save_table, from_select=ast_query, is_replace=True)

    else:
        raise NotImplementedError

    return ast_query


class Responce(Responder):
    when = {'aggregate': helpers.is_true}

    def result(self, query, request_env, mindsdb_env, session):
        db = query['$db']
        collection = query['aggregate']

        first_step = query['pipeline'][0]
        if '$match' in first_step:
            ast_query = aggregate_to_ast(query, request_env.get('database', 'mindsdb'))

            data = run_sql_command(request_env, ast_query)

        elif '$collStats' in first_step:
            raise ValueError(
                "To describe model use:"
                " db.runCommand({describe: 'model_name.attribute'})"
            )

        else:
            raise NotImplementedError

        cursor = {
            'id': Int64(0),
            'ns': f"{db}.$cmd.{collection}",
            'firstBatch': data
        }
        return {
            'cursor': cursor,
            'ok': 1
        }


responder = Responce()



import pandas as pd

from mindsdb_sql.planner.steps import (
    LimitOffsetStep,
    DataStep,
)

from mindsdb.api.executor.sql_query.result_set import ResultSet

from .base import BaseStepCall


class LimitOffsetStepCall(BaseStepCall):

    bind = LimitOffsetStep

    def call(self, step):
        step_data = self.steps_data[step.dataframe.step_num]

        step_data2 = ResultSet()
        for col in step_data.columns:
            step_data2.add_column(col)

        records = step_data.get_records_raw()

        if isinstance(step.offset, int):
            records = records[step.offset:]
        if isinstance(step.limit, int):
            records = records[:step.limit]

        for record in records:
            step_data2.add_record_raw(record)

        return step_data2


class DataStepCall(BaseStepCall):

    bind = DataStep

    def call(self, step):
        # create resultset
        df = pd.DataFrame(step.data)
        return ResultSet().from_df(df, database='', table_name='')



import pandas as pd

from typing import List

from mindsdb.integrations.libs.api_handler import APITable
from mindsdb.integrations.utilities.sql_utils import extract_comparison_conditions
from mindsdb.utilities import log

from mindsdb_sql.parser import ast

logger = log.getLogger(__name__)

class GitlabIssuesTable(APITable):
    """The GitLab Issue Table implementation"""

    def select(self, query: ast.Select) -> pd.DataFrame:
        """Pulls data from the GitLab "List repository issues" API
        Args:
            query: SELECT
        Returns:
            DataFrame
        Raises:
            ValueError
        """

        conditions = extract_comparison_conditions(query.where)

        if query.limit:
            total_results = query.limit.value
        else:
            total_results = 20

        issues_kwargs = {}
        order_by_conditions = {}

        if query.order_by and len(query.order_by) > 0:
            order_by_conditions["columns"] = []
            order_by_conditions["ascending"] = []

            for an_order in query.order_by:
                if an_order.field.parts[1] in self.get_columns():
                    order_by_conditions["columns"].append(an_order.field.parts[1])

                    if an_order.direction == "ASC":
                        order_by_conditions["ascending"].append(True)
                    else:
                        order_by_conditions["ascending"].append(False)
                else:
                    raise ValueError(
                        f"Order by unknown column {an_order.field.parts[1]}"
                    )

        for a_where in conditions:
            if a_where[1] == "state":
                if a_where[0] != "=":
                    raise ValueError("Unsupported where operation for state")
                if a_where[2] not in ["opened", "closed", "all"]:
                    raise ValueError(
                        f"Unsupported where argument for state {a_where[2]}"
                    )

                issues_kwargs["state"] = a_where[2]

                continue
            if a_where[1] == "labels":
                if a_where[0] != "=":
                    raise ValueError("Unsupported where operation for state")

                issues_kwargs["labels"] = a_where[2].split(",")

                continue
            if a_where[1] in ["assignee", "creator"]:
                if a_where[0] != "=":
                    raise ValueError(f"Unsupported where operation for {a_where[1]}")

                issues_kwargs[a_where[1]] = a_where[2]
            else:
                raise ValueError(f"Unsupported where argument {a_where[1]}")

        self.handler.connect()

        gitlab_issues_df = pd.DataFrame(columns=self.get_columns())

        start = 0

        issues_kwargs["per_page"] = total_results
        issues_kwargs["get_all"] = False
        while True:
            try:
                for issue in self.handler.connection.projects.get(
                    self.handler.repository
                ).issues.list(**issues_kwargs):

                    logger.debug(f"Processing issue {issue.iid}")

                    gitlab_issues_df = pd.concat(
                        [
                            gitlab_issues_df,
                            pd.DataFrame(
                                [
                                    {
                                        "number": issue.iid,
                                        "title": issue.title,
                                        "state": issue.state,
                                        "creator": issue.author["name"],
                                        "closed_by": issue.closed_by
                                        if issue.closed_by
                                        else None,
                                        "labels": ",".join(
                                            [label for label in issue.labels]
                                        ),
                                        "assignees": ",".join(
                                            [
                                                assignee["name"]
                                                for assignee in issue.assignees
                                            ]
                                        ),
                                        "body": issue.description,
                                        "created": issue.created_at,
                                        "updated": issue.updated_at,
                                        "closed": issue.closed_at,
                                    }
                                ]
                            ),
                        ]
                    )

                    if gitlab_issues_df.shape[0] >= total_results:
                        break
            except IndexError:
                break

            if gitlab_issues_df.shape[0] >= total_results:
                break
            else:
                break

        selected_columns = []
        for target in query.targets:
            if isinstance(target, ast.Star):
                selected_columns = self.get_columns()
                break
            elif isinstance(target, ast.Identifier):
                selected_columns.append(target.parts[-1])
            else:
                raise ValueError(f"Unknown query target {type(target)}")

        if len(gitlab_issues_df) == 0:
            gitlab_issues_df = pd.DataFrame([], columns=selected_columns)
        else:
            gitlab_issues_df.columns = self.get_columns()
            for col in set(gitlab_issues_df.columns).difference(set(selected_columns)):
                gitlab_issues_df = gitlab_issues_df.drop(col, axis=1)

            if len(order_by_conditions.get("columns", [])) > 0:
                gitlab_issues_df = gitlab_issues_df.sort_values(
                    by=order_by_conditions["columns"],
                    ascending=order_by_conditions["ascending"],
                )

        return gitlab_issues_df

    def get_columns(self) -> List[str]:
        """Gets all columns to be returned in pandas DataFrame responses
        Returns
        -------
        List[str]: list of columns
        """

        return [
            "number",
            "title",
            "state",
            "creator",
            "closed_by",
            "labels",
            "assignees",
            "body",
            "created",
            "updated",
            "closed",
        ]


class GitlabMergeRequestsTable(APITable):
    """The GitLab Merge Requests Table implementation"""

    def select(self, query: ast.Select) -> pd.DataFrame:
        """Pulls data from the GitLab "List repository rerge requests" API
        Args:
            query: SELECT
        Returns:
            DataFrame
        Raises:
            ValueError
        """

        conditions = extract_comparison_conditions(query.where)

        if query.limit:
            total_results = query.limit.value
        else:
            total_results = 20

        merge_requests_kwargs = {}
        order_by_conditions = {}

        if query.order_by and len(query.order_by) > 0:
            order_by_conditions["columns"] = []
            order_by_conditions["ascending"] = []

            for an_order in query.order_by:
                if an_order.field.parts[1] in self.get_columns():
                    order_by_conditions["columns"].append(an_order.field.parts[1])

                    if an_order.direction == "ASC":
                        order_by_conditions["ascending"].append(True)
                    else:
                        order_by_conditions["ascending"].append(False)
                else:
                    raise ValueError(
                        f"Order by unknown column {an_order.field.parts[1]}"
                    )

        for a_where in conditions:
            if a_where[1] == "state":
                if a_where[0] != "=":
                    raise ValueError("Unsupported where operation for state")
                if a_where[2] not in ["opened", "closed", "merged", "all"]:
                    raise ValueError(
                        f"Unsupported where argument for state {a_where[2]}"
                    )

                merge_requests_kwargs["state"] = a_where[2]

                continue
            if a_where[1] == "labels":
                if a_where[0] != "=":
                    raise ValueError("Unsupported where operation for labels")

                merge_requests_kwargs["labels"] = a_where[2].split(",")

                continue
            if a_where[1] in ["target_branch", "source_branch"]:
                if a_where[0] != "=":
                    raise ValueError(f"Unsupported where operation for {a_where[1]}")

                merge_requests_kwargs[a_where[1]] = a_where[2]
            else:
                raise ValueError(f"Unsupported where argument {a_where[1]}")

        self.handler.connect()

        gitlab_merge_requests_df = pd.DataFrame(columns=self.get_columns())

        start = 0

        merge_requests_kwargs["per_page"] = total_results
        merge_requests_kwargs["get_all"] = False
        while True:
            try:
                for merge_request in self.handler.connection.projects.get(
                    self.handler.repository
                ).mergerequests.list(**merge_requests_kwargs):

                    logger.debug(f"Processing merge request {merge_request.iid}")

                    gitlab_merge_requests_df = pd.concat(
                        [
                            gitlab_merge_requests_df,
                            pd.DataFrame(
                                [
                                    {
                                        "number": merge_request.iid,
                                        "title": merge_request.title,
                                        "state": merge_request.state,
                                        "creator": merge_request.author["name"],
                                        "closed_by": merge_request.closed_by
                                        if merge_request.closed_by
                                        else None,
                                        "mergeed_by": merge_request.merge_user["name"]
                                        if merge_request.merge_user
                                        else None,
                                        "labels": ",".join(
                                            [label for label in merge_request.labels]
                                        ),
                                        "assignees": ",".join(
                                            [
                                                assignee["name"]
                                                for assignee in merge_request.assignees
                                            ]
                                        ),
                                        "reviewers": ",".join(
                                            [
                                                reviewer["name"]
                                                for reviewer in merge_request.reviewers
                                            ]
                                        ),
                                        "body": merge_request.description,
                                        "target_branch": merge_request.target_branch,
                                        "source_branch": merge_request.source_branch,
                                        "upvotes": merge_request.upvotes,
                                        "downvotes": merge_request.downvotes,
                                        "draft": merge_request.draft,
                                        "work_in_progress": merge_request.work_in_progress,
                                        "milestone": merge_request.milestone["state"]
                                        if merge_request.milestone
                                        else None,
                                        "merge_status": merge_request.merge_status,
                                        "detailed_merge_status": merge_request.detailed_merge_status,
                                        "user_notes_count": merge_request.user_notes_count,
                                        "has_conflicts": merge_request.has_conflicts,
                                        "blocking_discussions_resolved": merge_request.blocking_discussions_resolved,
                                        "created": merge_request.created_at,
                                        "updated": merge_request.updated_at,
                                        "closed": merge_request.closed_at,
                                        "merged": merge_request.merged_at,
                                    }
                                ]
                            ),
                        ]
                    )

                    if gitlab_merge_requests_df.shape[0] >= total_results:
                        break
            except IndexError:
                break

            if gitlab_merge_requests_df.shape[0] >= total_results:
                break
            else:
                break

        selected_columns = []
        for target in query.targets:
            if isinstance(target, ast.Star):
                selected_columns = self.get_columns()
                break
            elif isinstance(target, ast.Identifier):
                selected_columns.append(target.parts[-1])
            else:
                raise ValueError(f"Unknown query target {type(target)}")

        if len(gitlab_merge_requests_df) == 0:
            gitlab_merge_requests_df = pd.DataFrame([], columns=selected_columns)
        else:
            gitlab_merge_requests_df.columns = self.get_columns()
            for col in set(gitlab_merge_requests_df.columns).difference(set(selected_columns)):
                gitlab_merge_requests_df = gitlab_merge_requests_df.drop(col, axis=1)

            if len(order_by_conditions.get("columns", [])) > 0:
                gitlab_merge_requests_df = gitlab_merge_requests_df.sort_values(
                    by=order_by_conditions["columns"],
                    ascending=order_by_conditions["ascending"],
                )

        return gitlab_merge_requests_df

    def get_columns(self) -> List[str]:
        """Gets all columns to be returned in pandas DataFrame responses
        Returns
        -------
        List[str]: list of columns
        """

        return [
            "number",
            "title",
            "state",
            "creator",
            "closed_by",
            "mergeed_by",
            "labels",
            "assignees",
            "reviewers",
            "body",
            "target_branch",
            "source_branch",
            "upvotes",
            "downvotes",
            "draft",
            "work_in_progress",
            "milestone",
            "merge_status",
            "detailed_merge_status",
            "user_notes_count",
            "has_conflicts",
            "blocking_discussions_resolved",
            "created",
            "updated",
            "closed",
            "merged",
        ]



from collections import OrderedDict

from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE


connection_args = OrderedDict(
    hosts={
        'type': ARG_TYPE.STR,
        'description': 'The host name(s) or IP address(es) of the Elasticsearch server(s). If multiple host name(s) or '
                       'IP address(es) exist, they should be separated by commas. This parameter is '
                       'optional, but it should be provided if cloud_id is not.'
    },
    cloud_id={
        'type': ARG_TYPE.STR,
        'description': 'The unique ID to your hosted Elasticsearch cluster on Elasticsearch Service. This parameter is '
                       'optional, but it should be provided if hosts is not.'
    },
    username={
        'type': ARG_TYPE.STR,
        'description': 'The user name used to authenticate with the Elasticsearch server. This parameter is optional.'
    },
    password={
        'type': ARG_TYPE.PWD,
        'description': 'The password to authenticate the user with the Elasticsearch server. This parameter is '
                       'optional.',
        'secret': True
    }
)

connection_args_example = OrderedDict(
    hosts='localhost:9200',
)



from http import HTTPStatus

from flask import request
from flask_restx import Resource, abort
from sqlalchemy.exc import NoResultFound


from mindsdb.api.http.namespaces.configs.projects import ns_conf
from mindsdb.api.executor.controllers.session_controller import SessionController
from mindsdb.metrics.metrics import api_endpoint_metrics


@ns_conf.route('/<project_name>/views')
class ViewsList(Resource):
    @ns_conf.doc('list_views')
    @api_endpoint_metrics('GET', '/views')
    def get(self, project_name):
        '''List all views'''
        session = SessionController()
        try:
            project = session.database_controller.get_project(project_name)
        except NoResultFound:
            abort(HTTPStatus.NOT_FOUND, f'Project name {project_name} does not exist')

        all_views = project.get_views()
        all_view_objs = []
        # Only want to return relevant fields to the user.
        for view in all_views:
            all_view_objs.append({
                'id': view['metadata']['id'],
                'name': view['name'],
                'query': view['query']
            })
        return all_view_objs

    @ns_conf.doc('create_view')
    @api_endpoint_metrics('POST', '/views')
    def post(self, project_name):
        '''Create a new view'''
        if 'view' not in request.json:
            abort(HTTPStatus.BAD_REQUEST, 'Must provide "view" parameter in POST body')
        session = SessionController()
        view_obj = request.json['view']
        if 'name' not in view_obj:
            abort(HTTPStatus.BAD_REQUEST, 'Missing "name" field for view')
        if 'query' not in view_obj:
            abort(HTTPStatus.BAD_REQUEST, 'Missing "query" field for view')
        name = view_obj['name']
        query = view_obj['query']

        try:
            project = session.database_controller.get_project(project_name)
        except NoResultFound:
            abort(HTTPStatus.NOT_FOUND, f'Project name {project_name} does not exist')

        if project.get_view(name) is not None:
            abort(HTTPStatus.CONFLICT, f'View with name {name} already exists.')

        project.create_view(name, query)
        created_view = project.get_view(name)
        # Only want to return relevant fields to the user.
        return {
            'id': created_view['metadata']['id'],
            'name': created_view['name'],
            'query': created_view['query']
        }, HTTPStatus.CREATED


@ns_conf.route('/<project_name>/views/<view_name>')
@ns_conf.param('project_name', 'Name of the project')
@ns_conf.param('view_name', 'Name of the view')
class ViewResource(Resource):
    @ns_conf.doc('get_view')
    @api_endpoint_metrics('GET', '/views/view')
    def get(self, project_name, view_name):
        '''Get a view by name'''
        session = SessionController()
        try:
            project = session.database_controller.get_project(project_name)
        except NoResultFound:
            abort(HTTPStatus.NOT_FOUND, f'Project name {project_name} does not exist')

        view = project.get_view(view_name)
        if view is None:
            abort(HTTPStatus.NOT_FOUND, f'View with name {view_name} does not exist')

        # Only want to return relevant fields to the user.
        return {
            'id': view['metadata']['id'],
            'name': view['name'],
            'query': view['query']
        }

    @ns_conf.doc('update_view')
    @api_endpoint_metrics('PUT', '/views/view')
    def put(self, project_name, view_name):
        '''Updates or creates a view'''
        if 'view' not in request.json:
            abort(HTTPStatus.BAD_REQUEST, 'Must provide "view" parameter in PUT body')
        request_view = request.json['view']
        session = SessionController()
        try:
            project = session.database_controller.get_project(project_name)
        except NoResultFound:
            abort(HTTPStatus.NOT_FOUND, f'Project name {project_name} does not exist')

        existing_view = project.get_view(view_name)
        if existing_view is None:
            # Create
            if 'query' not in request_view:
                abort(HTTPStatus.BAD_REQUEST, 'Missing "query" field for new view')
            project.create_view(view_name, request_view['query'])
            created_view = project.get_view(view_name)
            # Only want to return relevant fields to the user.
            return {
                'id': created_view['metadata']['id'],
                'name': created_view['name'],
                'query': created_view['query']
            }, HTTPStatus.CREATED

        new_query = existing_view['query']
        if 'query' in request_view:
            new_query = request_view['query']
            project.update_view(view_name, new_query)

        existing_view = project.get_view(view_name)
        # Only want to return relevant fields to the user.
        return {
            'id': existing_view['metadata']['id'],
            'name': existing_view['name'],
            'query': existing_view['query']
        }

    @ns_conf.doc('delete_view')
    @api_endpoint_metrics('DELETE', '/views/view')
    def delete(self, project_name, view_name):
        '''Deletes a view by name'''
        session = SessionController()
        try:
            project = session.database_controller.get_project(project_name)
        except NoResultFound:
            abort(HTTPStatus.NOT_FOUND, f'Project name {project_name} does not exist')

        if project.get_view(view_name) is None:
            abort(HTTPStatus.NOT_FOUND, f'View with name {view_name} does not exist')

        project.delete_view(view_name)
        return '', HTTPStatus.NO_CONTENT



import json
from flask import request

import msal
from ..exceptions import AuthException

from mindsdb.integrations.utilities.handlers.api_utilities import MSGraphAPIBaseClient

from mindsdb.utilities import log

logger = log.getLogger(__name__)


class MSGraphAPIAuthManager:
    def __init__(self, handler_storage: str, scopes: list, client_id: str, client_secret: str, tenant_id: str, code: str = None) -> None:
        self.handler_storage = handler_storage
        self.scopes = scopes
        self.client_id = client_id
        self.client_secret = client_secret
        self.tenant_id = tenant_id
        self.code = code

        # get the redirect uri from handler storage if it exists
        if self.handler_storage.json_get('args'):
            self.redirect_uri = self.handler_storage.json_get('args').get('redirect_uri')
        # otherwise, get it from the request headers
        # this is done because when the chatbot task runs, it doesn't have access to the request headers because it does not come through a request
        else:
            self.redirect_uri = request.headers['ORIGIN'] + '/verify-auth'
            if '127.0.0.1' in self.redirect_uri:
                self.redirect_uri = self.redirect_uri.replace('127.0.0.1', 'localhost')

            self.handler_storage.json_set('args', {'redirect_uri': self.redirect_uri})

    def get_access_token(self):
        try:
            creds = json.loads(self.handler_storage.file_get('creds'))
            access_token = creds.get('access_token')

            if not self._check_access_token_validity(access_token):
                logger.info('Access token expired. Refreshing...')
                response = self._refresh_access_token(creds.get('refresh_token'))
                self.handler_storage.file_set('creds', json.dumps(response).encode('utf-8'))
                access_token = response.get('access_token')

            return access_token
        except Exception as e:
            logger.error(f'Error getting credentials from storage: {e}!')

        response = self._execute_ms_graph_api_auth_flow()
        if "access_token" in response:
            self.handler_storage.file_set('creds', json.dumps(response).encode('utf-8'))
            return response['access_token']
        else:
            raise AuthException(
                f'Error getting access token: {response.get("error_description")}',
                auth_url=response.get('auth_url')
            )

    def _get_msal_app(self):
        return msal.ConfidentialClientApplication(
            self.client_id,
            authority=f"https://login.microsoftonline.com/{self.tenant_id}",
            client_credential=self.client_secret,
        )

    def _save_credentials_to_file(self, creds, creds_file):
        with open(creds_file, 'w') as f:
            f.write(json.dumps(creds))

    def _execute_ms_graph_api_auth_flow(self):
        msal_app = self._get_msal_app()

        if self.code:
            response = msal_app.acquire_token_by_authorization_code(
                code=self.code,
                scopes=self.scopes,
                redirect_uri=self.redirect_uri,
            )

            return response
        else:
            auth_url = msal_app.get_authorization_request_url(
                scopes=self.scopes,
                redirect_uri=self.redirect_uri,
            )

            raise AuthException(f'Authorisation required. Please follow the url: {auth_url}', auth_url=auth_url)

    def _refresh_access_token(self, refresh_token: str):
        msal_app = self._get_msal_app()

        response = msal_app.acquire_token_by_refresh_token(
            refresh_token=refresh_token,
            scopes=self.scopes,
        )

        return response

    def _check_access_token_validity(self, access_token: str):
        msal_graph_api_client = MSGraphAPIBaseClient(access_token)
        try:
            msal_graph_api_client.get_user_profile()
            return True
        except Exception as e:
            if 'InvalidAuthenticationToken' in str(e):
                return False
            else:
                raise e



import importlib
import time
from unittest.mock import patch
import pandas as pd
import pytest

from mindsdb_sql import parse_sql
from tests.unit.executor_test_base import BaseExecutorTest

try:
    importlib.import_module("pycaret")
    PYCARET_INSTALLED = True
except ImportError:
    PYCARET_INSTALLED = False


@pytest.mark.skipif(not PYCARET_INSTALLED, reason="pycaret is not installed")
class TestPyCaret(BaseExecutorTest):

    def wait_predictor(self, project, name):
        done = False
        for attempt in range(200):
            ret = self.run_sql(
                f"select * from {project}.models where name='{name}'"
            )
            if not ret.empty:
                if ret['STATUS'][0] == 'complete':
                    done = True
                    break
                elif ret['STATUS'][0] == 'error':
                    break
            time.sleep(0.5)
        if not done:
            raise RuntimeError("predictor wasn't created")

    def run_sql(self, sql):
        ret = self.command_executor.execute_command(
            parse_sql(sql, dialect='mindsdb')
        )
        assert ret.error_code is None
        if ret.data is not None:
            columns = [
                col.alias if col.alias is not None else col.name
                for col in ret.columns
            ]
            return pd.DataFrame(ret.data, columns=columns)

    @patch('mindsdb.integrations.handlers.postgres_handler.Handler')
    def test_classifier(self, mock_handler):
        df = pd.DataFrame({
            'sepal_length': [5.1, 4.9, 4.7, 4.6, 6.4, 6.9, 5.5, 6.5, 7.7, 6.3, 6.7, 7.2],
            'sepal_width': [3.5, 3.0, 3.2, 3.1, 3.2, 3.1, 2.3, 2.8, 2.8, 2.7, 3.3, 3.2],
            'petal_length': [1.4, 4.0, 1.3, 1.5, 4.5, 4.9, 4.0, 4.6, 6.7, 4.9, 5.7, 6.0],
            'petal_width': [0.2, 0.2, 0.2, 0.2, 1.5, 1.5, 1.3, 1.5, 2.0, 1.8, 2.1, 1.8],
            'species': ['Iris-setosa', 'Iris-setosa', 'Iris-setosa', 'Iris-setosa', 'Iris-versicolor', 'Iris-versicolor', 'Iris-versicolor', 'Iris-versicolor', 'Iris-virginica', 'Iris-virginica', 'Iris-virginica', 'Iris-virginica']
        })
        self.set_handler(mock_handler, name='pg', tables={'iris': df})

        # create project
        self.run_sql('create database proj;')

        # create predictor
        self.run_sql('''
            CREATE MODEL proj.my_pycaret_class_model
            FROM pg
                (SELECT sepal_length, sepal_width, petal_length, petal_width, species FROM iris)
            PREDICT species
            USING
              engine = 'pycaret',
              model_type = 'classification',
              model_name = 'xgboost',
              setup_session_id = 123,
              setup_fold = 2;
        ''')
        self.wait_predictor('proj', 'my_pycaret_class_model')

        # run predict
        ret = self.run_sql('''
            SELECT prediction_label
            FROM pg.iris as t
            JOIN proj.my_pycaret_class_model AS m;
        ''')

        assert ret['prediction_label'].iloc[0] == 'Iris-setosa'

    @patch('mindsdb.integrations.handlers.postgres_handler.Handler')
    def test_regression(self, mock_handler):
        df = pd.DataFrame({
            'age': [19, 18, 28, 33, 32, 31, 46, 37],
            'sex': ['female', 'male', 'male', 'male', 'male', 'female', 'female', 'female'],
            'bmi': [27.9, 33.77, 33, 22.705, 28.88, 25.74, 33.44, 27.74],
            'children': [0, 1, 3, 0, 0, 0, 1, 3],
            'smoker': ['yes', 'no', 'no', 'no', 'no', 'no', 'no', 'no'],
            'region': ['southwest', 'southeast', 'southeast', 'northwest', 'northwest', 'southeast', 'southeast', 'northwest'],
            'charges': [16884.924, 1725.5523, 4449.462, 21984.47061, 3866.8552, 3756.6216, 8240.5896, 7281.5056]
        })

        self.set_handler(mock_handler, name='pg', tables={'insurance': df})

        # create project
        self.run_sql('create database proj;')

        # create predictor
        self.run_sql('''
            CREATE MODEL proj.my_pycaret_regr_model
            FROM pg
                (SELECT age, sex, bmi, children, smoker, region, charges FROM insurance)
            PREDICT charges
            USING
              engine = 'pycaret',
              model_type = 'regression',
              model_name = 'xgboost',
              setup_session_id = 123,
              setup_fold = 2;
        ''')
        self.wait_predictor('proj', 'my_pycaret_regr_model')

        # run predict
        ret = self.run_sql('''
            SELECT prediction_label
            FROM pg.insurance as t
            JOIN proj.my_pycaret_regr_model AS m;
        ''')

        assert int(ret['prediction_label'].iloc[0]) == 3822

    @patch('mindsdb.integrations.handlers.postgres_handler.Handler')
    @pytest.mark.skip(reason="MindsDB recognizes 'Anomaly' as a keyword so it fails to fetch Anomaly column")
    def test_anomaly(self, mock_handler):
        df = pd.DataFrame({
            'Col1': [0.263995357, 0.764928588, 0.13842355, 0.935242061, 0.605866573, 0.518789697, 0.912225161, 0.608234451, 0.723781923, 0.73359095],
            'Col2': [0.546092303, 0.65397459, 0.065575135, 0.227771913, 0.845269445, 0.837065879, 0.272378939, 0.331678698, 0.429296975, 0.367422001],
            'Col3': [0.336714104, 0.538842451, 0.192801069, 0.553562822, 0.074514511, 0.332993162, 0.365792205, 0.861309323, 0.899016587, 0.088600152],
            'Col4': [0.092107835, 0.995016662, 0.014465045, 0.176370646, 0.241530075, 0.514723634, 0.562208164, 0.158963258, 0.073715215, 0.208463224],
            'Col5': [0.325261175, 0.805967636, 0.957033424, 0.331664957, 0.307923366, 0.355314772, 0.50189852, 0.558449452, 0.885169295, 0.182754409],
            'Col6': [0.212464853, 0.780304761, 0.458443656, 0.634508561, 0.373030452, 0.465650668, 0.413997158, 0.013080054, 0.570250227, 0.736672363],
            'Col7': [0.258565714, 0.437317789, 0.559647989, 0.109202597, 0.994553306, 0.896994183, 0.488468506, 0.251942977, 0.017265143, 0.538513303],
            'Col8': [0.869236755, 0.277978893, 0.42307639, 0.11247202, 0.183727053, 0.034959735, 0.111113968, 0.249329646, 0.550683376, 0.049843054],
            'Col9': [0.197077957, 0.843918225, 0.24339588, 0.281278233, 0.329148141, 0.73458152, 0.191947043, 0.927804425, 0.71326865, 0.891548497],
            'Col10': [0.292984504, 0.70343162, 0.43962138, 0.107867968, 0.922947409, 0.25345779, 0.29565178, 0.355286799, 0.980911322, 0.308864217]
        })

        self.set_handler(mock_handler, name='pg', tables={'anomaly': df})

        # create project
        self.run_sql('create database proj;')

        # create predictor
        self.run_sql('''
            CREATE MODEL proj.my_pycaret_anom_model
            FROM pg
                (SELECT Col1, Col2, Col3, Col4, Col5, Col6, Col7, Col8, Col9, Col10 FROM anomaly)
            PREDICT Col10
            USING
              engine = 'pycaret',
              model_type = 'anomaly',
              model_name = 'iforest',
              setup_session_id = 123,
              setup_fold = 2;
        ''')
        self.wait_predictor('proj', 'my_pycaret_anom_model')

        # run predict
        # TODO: is there a workaround for this? (it works when ran in web UI)
        ret = self.run_sql('''
            SELECT m.Anomaly
            FROM pg.anomaly as t
            JOIN proj.my_pycaret_anom_model AS m;
        ''')

        assert int(ret['Anomaly'].iloc[0]) == 0

    @patch('mindsdb.integrations.handlers.postgres_handler.Handler')
    def test_cluster(self, mock_handler):
        df = pd.DataFrame({
            'Age': [58, 59, 62, 59, 87, 29, 54, 87],
            'Income': [77769, 81799, 74751, 74373, 17760, 13157, 76500, 42592],
            'SpendingScore': [0.7913287771988531, 0.7910820467274178, 0.7026569520102857, 0.7656795619984281, 0.3487775484305076, 0.8470341025128374, 0.7851978501165687, 0.3552896820382753],
            'Savings': [6559.8299230048315, 5417.661426197439, 9258.992965034067, 7346.334503537976, 16869.507130301474, 3535.5143522162816, 6878.884248553975, 18086.287157859304]
        })

        self.set_handler(mock_handler, name='pg', tables={'jewellery': df})

        # create project
        self.run_sql('create database proj;')

        # create predictor
        self.run_sql('''
            CREATE MODEL proj.my_pycaret_cluster_model
            FROM pg
                (SELECT Age, Income, SpendingScore, Savings FROM jewellery)
            PREDICT Savings
            USING
              engine = 'pycaret',
              model_type = 'clustering',
              model_name = 'kmeans',
              setup_session_id = 123;
        ''')
        self.wait_predictor('proj', 'my_pycaret_cluster_model')

        # run predict
        ret = self.run_sql('''
            SELECT m.Cluster
            FROM pg.jewellery as t
            JOIN proj.my_pycaret_cluster_model AS m;
        ''')

        assert ret['Cluster'].iloc[0] == "Cluster 0"

    @patch('mindsdb.integrations.handlers.postgres_handler.Handler')
    def test_timeseries(self, mock_handler):
        df = pd.DataFrame({
            'Year': [1949, 1949, 1949, 1949, 1949, 1949, 1949, 1949, 1949, 1949, 1949, 1949, 1950, 1950, 1950, 1950, 1950, 1950, 1950, 1950],
            'Month': [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 1, 2, 3, 4, 5, 6, 7, 8],
            'Passengers': [112, 118, 132, 129, 121, 135, 148, 148, 136, 119, 104, 118, 115, 126, 141, 135, 125, 149, 170, 170]
        })

        self.set_handler(mock_handler, name='pg', tables={'airline': df})

        # create project
        self.run_sql('create database proj;')

        # create predictor
        self.run_sql('''
            CREATE MODEL proj.my_pycaret_timeseries_model
            FROM pg
                (SELECT Year, Month, Passengers FROM airline)
            PREDICT Passengers
            USING
              engine = 'pycaret',
              model_type = 'time_series',
              model_name = 'naive',
              setup_fh = 3,
              predict_fh = 36,
              setup_session_id = 123;
        ''')
        self.wait_predictor('proj', 'my_pycaret_timeseries_model')

        # run predict
        ret = self.run_sql('''
            SELECT m.y_pred
            FROM pg.airline as t
            JOIN proj.my_pycaret_timeseries_model AS m;
        ''')

        assert int(ret['y_pred'].iloc[0]) == 125



from mindsdb.integrations.handlers.postgres_handler import Handler as PostgresHandler


class QuestDBHandler(PostgresHandler):
    """
    This handler handles connection and execution of the QuestDB statements. 
    TODO: check the dialect for questdb
    """
    name = 'questdb'

    def __init__(self, name, **kwargs):
        super().__init__(name, **kwargs)

    def get_tables(self):
        """
        List all tabels in QuestDB
        """
        query = "SHOW TABLES"
        response = super().native_query(query)
        return response

    def get_columns(self, table_name):
        """
        List information about the table
        """
        query = f"SELECT * FROM tables() WHERE name='{table_name}';"
        response = super().native_query(query)
        return response



from collections import OrderedDict

from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE


connection_args = OrderedDict(
    db_url={
        "type": ARG_TYPE.STR,
        "description": "Xata database url with region, database and, optionally the branch information",
        "required": True,
    },
    api_key={
        "type": ARG_TYPE.STR,
        "description": "personal Xata API key",
        "required": True,
        "secret": True
    },
    dimension={
        "type": ARG_TYPE.INT,
        "description": "default dimension of embeddings vector used to create table when using create (default=8)",
        "required": False,
    },
    similarity_function={
        "type": ARG_TYPE.STR,
        "description": "similarity function to use for vector searches (default=cosineSimilarity)",
        "required": False,
    }
)

connection_args_example = OrderedDict(
    db_url="https://...",
    api_key="abc_def...",
    dimension=8,
    similarity_function="l1"
)



from mindsdb_sql.parser.ast import (
    Identifier,
)
from mindsdb_sql.planner.steps import (
    SaveToTable,
    InsertToTable,
    CreateTableStep
)

from mindsdb.api.executor.sql_query.result_set import ResultSet, Column
from mindsdb.api.executor.exceptions import (
    NotSupportedYet,
    LogicError
)

from .base import BaseStepCall


class InsertToTableCall(BaseStepCall):

    bind = InsertToTable

    def call(self, step):
        is_replace = False
        is_create = False

        if type(step) == SaveToTable:
            is_create = True

            if step.is_replace:
                is_replace = True

        if step.dataframe is not None:
            data = step.dataframe.result_data
        elif step.query is not None:
            data = ResultSet()
            for col in step.query.columns:
                data.add_column(Column(col.name))

            for row in step.query.values:
                record = [v.value for v in row]
                data.add_record_raw(record)
        else:
            raise LogicError(f'Data not found for insert: {step}')

        if len(step.table.parts) > 1:
            integration_name = step.table.parts[0]
            table_name = Identifier(parts=step.table.parts[1:])
        else:
            integration_name = self.context['database']
            table_name = step.table

        dn = self.session.datahub.get(integration_name)

        if hasattr(dn, 'create_table') is False:
            raise NotSupportedYet(f"Creating table in '{integration_name}' is not supported")

        #  del 'service' columns
        for col in data.find_columns('__mindsdb_row_id'):
            data.del_column(col)
        for col in data.find_columns('__mdb_forecast_offset'):
            data.del_column(col)

        # region del columns filtered at projection step
        columns_list = self.get_columns_list()
        if columns_list is not None:
            filtered_column_names = [x.name for x in columns_list]
            for col in data.columns:
                if col.name.startswith('predictor.'):
                    continue
                if col.name in filtered_column_names:
                    continue
                data.del_column(col)
        # endregion

        # drop double names
        col_names = set()
        for col in data.columns:
            if col.alias in col_names:
                data.del_column(col)
            else:
                col_names.add(col.alias)

        dn.create_table(
            table_name=table_name,
            result_set=data,
            is_replace=is_replace,
            is_create=is_create
        )
        return ResultSet()


class SaveToTableCall(InsertToTableCall):

    bind = SaveToTable


class CreateTableCall(BaseStepCall):

    bind = CreateTableStep

    def call(self, step):

        if len(step.table.parts) > 1:
            integration_name = step.table.parts[0]
            table_name = Identifier(parts=step.table.parts[1:])
        else:
            integration_name = self.context['database']
            table_name = step.table

        dn = self.session.datahub.get(integration_name)

        dn.create_table(
            table_name=table_name,
            columns=step.columns,
            is_replace=step.is_replace,
            is_create=True
        )
        return ResultSet()



from collections import OrderedDict

from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE


connection_args = OrderedDict(
    url={
        'type': ARG_TYPE.STR,
        'description': 'The URI-Like connection string to the MariaDB server. If provided, it will override the other connection arguments.',
        'required': False,
        'label': 'URL'
    },
    user={
        'type': ARG_TYPE.STR,
        'description': 'The user name used to authenticate with the MariaDB server.',
        'required': True,
        'label': 'User'
    },
    password={
        'type': ARG_TYPE.PWD,
        'description': 'The password to authenticate the user with the MariaDB server.',
        'required': True,
        'label': 'Password',
        'secret': True
    },
    database={
        'type': ARG_TYPE.STR,
        'description': 'The database name to use when connecting with the MariaDB server.',
        'required': True,
        'label': 'Database'
    },
    host={
        'type': ARG_TYPE.STR,
        'description': 'The host name or IP address of the MariaDB server. NOTE: use \'127.0.0.1\' instead of \'localhost\' to connect to local server.',
        'required': True,
        'label': 'Host'
    },
    port={
        'type': ARG_TYPE.INT,
        'description': 'The TCP/IP port of the MariaDB server. Must be an integer.',
        'required': True,
        'label': 'Port'
    },
    ssl={
        'type': ARG_TYPE.BOOL,
        'description': 'Set it to True to enable ssl.',
        'required': False,
        'label': 'ssl'
    },
    ssl_ca={
        'type': ARG_TYPE.PATH,
        'description': 'Path or URL of the Certificate Authority (CA) certificate file',
        'required': False,
        'label': 'ssl_ca'
    },
    ssl_cert={
        'type': ARG_TYPE.PATH,
        'description': 'Path name or URL of the server public key certificate file',
        'required': False,
        'label': 'ssl_cert'
    },
    ssl_key={
        'type': ARG_TYPE.PATH,
        'description': 'The path name or URL of the server private key file',
        'required': False,
        'label': 'ssl_key',
    }
)

connection_args_example = OrderedDict(
    host='127.0.0.1',
    port=3306,
    user='root',
    password='password',
    database='database'
)



from typing import Text, List, Dict, Optional
from mindsdb.integrations.utilities.handlers.api_utilities.microsoft.ms_graph_api_utilities import MSGraphAPIBaseClient


class MSGraphAPITeamsClient(MSGraphAPIBaseClient): 
    """
    The Microsoft Graph API client for the Microsoft Teams handler.
    This client is used for accessing the Microsoft Teams specific endpoints of the Microsoft Graph API.
    Several common methods for submitting requests, fetching data, etc. are inherited from the base class.
    """

    def _get_group_ids(self) -> List[Text]:
        """
        Get all group IDs related to Microsoft Teams.

        Returns
        -------
        List[Text]
            The group IDs.
        """

        if not self._group_ids:
            api_url = self._get_api_url("groups")
            # only get the id and resourceProvisioningOptions fields
            params = {"$select": "id,resourceProvisioningOptions"}
            groups = self._get_response_value_unsafe(self._make_request(api_url, params=params))
            # filter out only the groups that are related to Microsoft Teams
            self._group_ids = [item["id"] for item in groups if "Team" in item["resourceProvisioningOptions"]]

        return self._group_ids
    
    def _get_channel_ids(self, group_id: Text) -> List[Text]:
        """
        Get all channel IDs related to a group.

        Parameters
        ----------
        group_id : Text
            The ID of the group.

        Returns
        -------
        List[Text]
            The channel IDs for that group.
        """

        api_url = self._get_api_url(f"teams/{group_id}/channels")
        channels_ids = self._get_response_value_unsafe(self._make_request(api_url))

        return channels_ids
    
    def get_channel(self, group_id: Text, channel_id: Text) -> Dict:
        """
        Get a channel by its ID and the ID of the group that it belongs to.

        Parameters
        ----------
        group_id : str
            The ID of the group that the channel belongs to.

        channel_id : str
            The ID of the channel.

        Returns
        -------
        Dict
            The channel data.
        """

        api_url = self._get_api_url(f"teams/{group_id}/channels/{channel_id}")
        channel = self._make_request(api_url)
        # add the group ID to the channel data
        channel.update({"teamId": group_id})

        return channel
    
    def get_channels(self) -> List[Dict]:
        """
        Get all channels.

        Returns
        -------
        List[Dict]
            The channels data.
        """

        channels = []
        for group_id in self._get_group_ids():
            for group_channels in self._fetch_data(f"teams/{group_id}/channels", pagination=False):
                [group_channel.update({"teamId": group_id}) for group_channel in group_channels]
                channels.extend(group_channels)

        return channels
    
    def get_channel_message(self, group_id: Text, channel_id: Text, message_id: Text) -> Dict:
        """
        Get a channel message by its ID and the IDs of the group and channel that it belongs to.

        Parameters
        ----------
        group_id : str
            The ID of the group that the channel belongs to.

        channel_id : str
            The ID of the channel that the message belongs to.

        message_id : str
            The ID of the message.

        Returns
        -------
        Dict
            The channel message data.
        """

        api_url = self._get_api_url(f"teams/{group_id}/channels/{channel_id}/messages/{message_id}")
        message = self._make_request(api_url)

        return message

    def get_channel_messages(self) -> List[Dict]:
        """
        Get all channel messages.

        Returns
        -------
        List[Dict]
            The channel messages data.
        """

        channel_messages = []
        for group_id in self._get_group_ids():
            for channel_id in self._get_channel_ids(group_id):
                for messages in self._fetch_data(f"teams/{group_id}/channels/{channel_id['id']}/messages"):
                    channel_messages.extend(messages)

        return channel_messages
    
    def send_channel_message(self, group_id: Text, channel_id: Text, message: Text, subject: Optional[Text] = None) -> None:
        """
        Send a message to a channel.

        Parameters
        ----------
        group_id : Text
            The ID of the group that the channel belongs to.

        channel_id : Text
            The ID of the channel.

        message : Text
            The message to send.

        subject : Text, Optional
            The subject of the message.
        """

        api_url = self._get_api_url(f"teams/{group_id}/channels/{channel_id}/messages")
        data = {
            "subject": subject,
            "body": {
                "content": message
            }
        }

        self._make_request(api_url, data=data, method="POST")

    def get_chat(self, chat_id: Text) -> Dict:
        """
        Get a chat by its ID.

        Parameters
        ----------
        chat_id : str
            The ID of the chat.
        """

        api_url = self._get_api_url(f"chats/{chat_id}")
        # expand the response with the last message preview
        chat = self._make_request(api_url, params={"$expand": "lastMessagePreview"})

        return chat
    
    def get_chats(self) -> List[Dict]:
        """
        Get all chats.

        Returns
        -------
        List[Dict]
            The chats data.
        """

        chats = []
        for chat in self._fetch_data("chats", params={"$expand": "lastMessagePreview"}):
            chats.extend(chat)

        return chats
    
    def get_chat_message(self, chat_id: Text, message_id: Text) -> Dict:
        """
        Get a chat message by its ID and the ID of the chat that it belongs to.

        Parameters
        ----------
        chat_id : str
            The ID of the chat that the message belongs to.

        message_id : str
            The ID of the message.

        Returns
        -------
        Dict
            The chat message data.
        """

        api_url = self._get_api_url(f"chats/{chat_id}/messages/{message_id}")
        message = self._make_request(api_url)

        return message
    
    def get_chat_messages(self, chat_id: Text) -> List[Dict]:
        """
        Get all chat messages for a chat.

        Parameters
        ----------
        chat_id : str
            The ID of the chat.

        Returns
        -------
        List[Dict]
            The chat messages data.
        """

        chat_messages = []
        for messages in self._fetch_data(f"chats/{chat_id}/messages"):
            chat_messages.extend(messages)

        return chat_messages

    def get_all_chat_messages(self) -> List[Dict]:
        """
        Get all chat messages for all chats.

        Returns
        -------
        List[Dict]
            The chat messages data.
        """

        chat_messages = []
        for chat_id in [chat["id"] for chat in self.get_chats()]:
            for messages in self._fetch_data(f"chats/{chat_id}/messages"):
                chat_messages.extend(messages)

        return chat_messages
    
    def send_chat_message(self, chat_id: Text, message: Text, subject: Optional[Text] = None) -> None:
        """
        Send a message to a chat.

        Parameters
        ----------
        chat_id : Text
            The ID of the chat.

        message : Text
            The message to send.

        subject : Text, Optional
            The subject of the message.
        """

        api_url = self._get_api_url(f"chats/{chat_id}/messages")
        data = {
            "subject": subject,
            "body": {
                "content": message
            }
        }
        
        self._make_request(api_url, data=data, method="POST")



import copy

from mindsdb_sql.parser.ast import (
    BinaryOperation,
    UnaryOperation,
    Constant,
)
from mindsdb_sql.planner.steps import (
    MapReduceStep,
    FetchDataframeStep,
    MultipleSteps,
)

from mindsdb.api.executor.sql_query.result_set import ResultSet
from mindsdb.api.executor.exceptions import LogicError

from .base import BaseStepCall
from .fetch_dataframe import FetchDataframeStepCall


def markQueryVar(where):
    if isinstance(where, BinaryOperation):
        markQueryVar(where.args[0])
        markQueryVar(where.args[1])
    elif isinstance(where, UnaryOperation):
        markQueryVar(where.args[0])
    elif isinstance(where, Constant):
        if str(where.value).startswith('$var['):
            where.is_var = True
            where.var_name = where.value


def unmarkQueryVar(where):
    if isinstance(where, BinaryOperation):
        unmarkQueryVar(where.args[0])
        unmarkQueryVar(where.args[1])
    elif isinstance(where, UnaryOperation):
        unmarkQueryVar(where.args[0])
    elif isinstance(where, Constant):
        if hasattr(where, 'is_var') and where.is_var is True:
            where.value = where.var_name


def replaceQueryVar(where, var_value, var_name):
    if isinstance(where, BinaryOperation):
        replaceQueryVar(where.args[0], var_value, var_name)
        replaceQueryVar(where.args[1], var_value, var_name)
    elif isinstance(where, UnaryOperation):
        replaceQueryVar(where.args[0], var_value, var_name)
    elif isinstance(where, Constant):
        if hasattr(where, 'is_var') and where.is_var is True and where.value == f'$var[{var_name}]':
            where.value = var_value


def join_query_data(target, source):
    if len(target.columns) == 0:
        target = source
    else:
        target.add_records(source.get_records())
    return target


class MapReduceStepCall(BaseStepCall):

    bind = MapReduceStep

    def call(self, step):
        if step.reduce != 'union':
            raise LogicError(f'Unknown MapReduceStep type: {step.reduce}')

        step_data = self.steps_data[step.values.step_num]
        vars = []
        for row in step_data.get_records():
            var_group = {}
            vars.append(var_group)
            for name, value in row.items():
                if name != '__mindsdb_row_id':
                    var_group[name] = value

        data = ResultSet()

        substep = step.step
        if type(substep) is FetchDataframeStep:
            query = substep.query
            if len(vars) == 0:
                substep.query.limit = Constant(0)
                substep.query.where = None

                sub_data = self._fetch_dataframe_step(substep)

                for column in sub_data.columns:
                    data.add_column(column)

                data.add_records(sub_data.get_records())

            for var_group in vars:
                markQueryVar(query.where)
                for name, value in var_group.items():
                    replaceQueryVar(query.where, value, name)
                sub_data = self._fetch_dataframe_step(substep)
                if len(data.columns) == 0:
                    data = sub_data
                else:
                    data.add_records(sub_data.get_records())

                unmarkQueryVar(query.where)
        elif type(substep) is MultipleSteps:
            data = self._multiple_steps_reduce(substep, vars)
        else:
            raise LogicError(f'Unknown step type: {step.step}')
        return data

    def _multiple_steps_reduce(self, step, vars):
        if step.reduce != 'union':
            raise LogicError(f'Unknown MultipleSteps type: {step.reduce}')

        data = ResultSet()

        # mark vars
        steps = []
        for substep in step.steps:
            if isinstance(substep, FetchDataframeStep) is False:
                raise LogicError(f'Wrong step type for MultipleSteps: {step}')
            substep = copy.deepcopy(substep)
            markQueryVar(substep.query.where)
            steps.append(substep)

        for var_group in vars:
            steps2 = copy.deepcopy(steps)
            for name, value in var_group.items():
                for substep in steps2:
                    replaceQueryVar(substep.query.where, value, name)
            sub_data = self._multiple_steps(steps2)
            data = join_query_data(data, sub_data)

        return data

    def _multiple_steps(self, steps):
        data = ResultSet()
        for substep in steps:
            sub_data = self._fetch_dataframe_step(substep)
            data = join_query_data(data, sub_data)
        return data

    def _fetch_dataframe_step(self, step):
        return FetchDataframeStepCall(self.sql_query).call(step)



import sib_api_v3_sdk
import pandas as pd


from typing import List, Optional, Dict, Text, Any
from mindsdb.utilities import log
from mindsdb.integrations.libs.api_handler import APITable

from mindsdb_sql.parser import ast
from sib_api_v3_sdk.rest import ApiException
from datetime import datetime


from mindsdb.integrations.utilities.handlers.query_utilities import (
    SELECTQueryParser,
    SELECTQueryExecutor,
    UPDATEQueryExecutor,
    UPDATEQueryParser,
    DELETEQueryParser,
    DELETEQueryExecutor,
    INSERTQueryParser,
)

logger = log.getLogger(__name__)


class EmailCampaignsTable(APITable):
    """The Sendinblue Email Campaigns Table implementation"""

    def select(self, query: ast.Select) -> pd.DataFrame:
        """Pulls data from the Sendinblue "GET /emailCampaigns" API endpoint.

        Parameters
        ----------
        query : ast.Select
           Given SQL SELECT query

        Returns
        -------
        pd.DataFrame
            Sendinblue Email Campaigns matching the query

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """

        select_statement_parser = SELECTQueryParser(
            query, 'email_campaigns', self.get_columns()
        )
        (
            selected_columns,
            where_conditions,
            order_by_conditions,
            result_limit,
        ) = select_statement_parser.parse_query()

        email_campaigns_df = pd.json_normalize(
            self.get_email_campaigns(limit=result_limit)
        )

        select_statement_executor = SELECTQueryExecutor(
            email_campaigns_df, selected_columns, where_conditions, order_by_conditions
        )
        email_campaigns_df = select_statement_executor.execute_query()

        return email_campaigns_df

    def get_columns(self) -> List[str]:
        return pd.json_normalize(self.get_email_campaigns(limit=1)).columns.tolist()

    def get_email_campaigns(self, **kwargs):
        connection = self.handler.connect()
        email_campaigns_api_instance = sib_api_v3_sdk.EmailCampaignsApi(connection)
        email_campaigns = email_campaigns_api_instance.get_email_campaigns(**kwargs)
        return [email_campaign for email_campaign in email_campaigns.campaigns]

    def delete(self, query: ast.Delete) -> None:
        """
        Deletes an email campaign from Sendinblue.

        Parameters
        ----------
        query : ast.Delete
           Given SQL DELETE query

        Returns
        -------
        None

        Raises
        ------
        RuntimeError
            If an error occurs when calling Sendinblue's API
        """
        # this  parses the DELETE statement to extract where conditions
        delete_statement_parser = DELETEQueryParser(query)
        where_conditions = delete_statement_parser.parse_query()
        # this retrieves the current list of email campaigns and normalize the data into a DataFrame
        email_campaigns_df = pd.json_normalize(self.get_email_campaigns())
        # this execute the delete query  to filter out the campaigns to be deleted
        delete_query_executor = DELETEQueryExecutor(
            email_campaigns_df, where_conditions
        )
        # this gets the updated DataFrame after executing  delete conditions
        email_campaigns_df = delete_query_executor.execute_query()
        campaign_ids = email_campaigns_df['id'].tolist()
        self.delete_email_campaigns(campaign_ids)

    def delete_email_campaigns(self, campaign_ids: List[Text]) -> None:
        # this establish a connection to Sendinblue API
        connection = self.handler.connect()
        email_campaigns_api_instance = sib_api_v3_sdk.EmailCampaignsApi(connection)
        for campaign_id in campaign_ids:
            try:
                email_campaigns_api_instance.delete_email_campaign(campaign_id)
                logger.info(f'Email Campaign {campaign_id} deleted')
            except ApiException as e:
                logger.error(
                    f"Exception when calling EmailCampaignsApi->delete_email_campaign: {e}\n"
                )
                raise RuntimeError(
                    f"Failed to execute the delete command for Email Campaign {campaign_id}"
                ) from e

    def update(self, query: 'ast.Update') -> None:
        """
        Updates data in Sendinblue "PUT /emailCampaigns/{campaignId}" API endpoint.

        Parameters
        ----------
        query : ast.Update
            Given SQL UPDATE query

        Returns
        -------
        None

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        RuntimeError
            If an error occurs when calling Sendinblue's API
        """
        # this parse the UPDATE statement to extract the new values and the conditions for the update
        update_statement_parser = UPDATEQueryParser(query)
        values_to_update, where_conditions = update_statement_parser.parse_query()

        email_campaigns_df = pd.json_normalize(self.get_email_campaigns())
        update_query_executor = UPDATEQueryExecutor(
            email_campaigns_df, where_conditions
        )
        # this retrieves the current list of email campaigns
        email_campaigns_df = update_query_executor.execute_query()
        # this  extracts the IDs of the campaigns that have been updated
        campaign_ids = email_campaigns_df['id'].tolist()

        self.update_email_campaigns(campaign_ids, values_to_update)

    def update_email_campaigns(
        self, campaign_ids: List[int], values_to_update: Dict
    ) -> None:
        # this establish a connection to Sendinblue API

        connection = self.handler.connect()
        email_campaigns_api_instance = sib_api_v3_sdk.EmailCampaignsApi(connection)

        for campaign_id in campaign_ids:
            try:
                email_campaigns_api_instance.update_email_campaign(
                    campaign_id, values_to_update
                )
                logger.info(f'Email Campaign {campaign_id} updated')
            except ApiException as e:
                logger.error(
                    f"Exception when calling EmailCampaignsApi->update_email_campaign: {e}\n"
                )
                raise RuntimeError(
                    f"Failed to execute the update command for Email Campaign {campaign_id}"
                ) from e
    def insert(self, query: 'ast.Insert') -> None:
        """
        Inserts new email campaigns into Sendinblue.

        Parameters
        ----------
        query : ast.Insert
            The SQL INSERT query to be parsed and executed.

        Raises
        ------
        ValueError
            If the necessary sender information is incomplete or incorrectly formatted.
        Exception
            For any unexpected errors during the email campaign creation.
        """
        #this defines columns that are supported and mandatory for an INSERT operation.
        supported_columns = [
            'name', 'subject', 'sender_name', 'sender_email', 
            'html_content', 'scheduled_at', 'recipients_lists', 'tag'
        ]
        mandatory_columns = ['name', 'subject', 'sender_name', 'sender_email', 'html_content']

        #this Parse the INSERT query to extract data.
        insert_statement_parser = INSERTQueryParser(
            query, supported_columns=supported_columns, 
            mandatory_columns=mandatory_columns, all_mandatory=True
        )
        email_campaigns_data = insert_statement_parser.parse_query()

        #this processes each campaign data extracted from the query.
        for email_campaign_data in email_campaigns_data:
            #this extracts and format sender information.
            sender_info = {}
            if 'sender_name' in email_campaign_data:
                sender_info['name'] = email_campaign_data.pop('sender_name')
            if 'sender_email' in email_campaign_data and email_campaign_data['sender_email'] is not None:
                sender_info['email'] = email_campaign_data.pop('sender_email')
            if 'sender_id' in email_campaign_data and email_campaign_data['sender_id'] is not None:
                sender_info['id'] = email_campaign_data.pop('sender_id')

            # this validates sender information.
            if not sender_info.get('name') or (not sender_info.get('email') and not sender_info.get('id')):
                raise ValueError("Sender information is incomplete or incorrectly formatted.")

            email_campaign_data['sender'] = sender_info

            #this creates each email campaign.
            self.create_email_campaign(email_campaign_data)

    def create_email_campaign(self, email_campaign_data: Dict[str, Any]) -> None:
        """
        Creates a new email campaign in Sendinblue.

        Parameters
        ----------
        email_campaign_data : Dict[str, Any]
            The data for the email campaign to be created.

        Raises
        ------
        Exception
            For any errors during the email campaign creation process.
        """
        #this establish a connection to the Sendinblue API.
        api_session = self.handler.connect()
        email_campaigns_api_instance = sib_api_v3_sdk.EmailCampaignsApi(api_session)

        #this logs the data for the email campaign being created.
        logger.info(f"Email campaign data before creating the object: {email_campaign_data}")

        try:
            #this creates the email campaign object and send it to Sendinblue.
            email_campaign = sib_api_v3_sdk.CreateEmailCampaign(**email_campaign_data)
            logger.info(f"Email campaign object after creation: {email_campaign}")

            #this executes the API call to create the campaign.
            created_campaign = email_campaigns_api_instance.create_email_campaign(email_campaign)

            #this checks and log the response from the API.
            if 'id' not in created_campaign.to_dict():
                logger.error('Email campaign creation failed')
            else:
                logger.info(f'Email Campaign {created_campaign.to_dict()["id"]} created')
        except ApiException as e:
            #this handles API exceptions and log the detailed response.
            logger.error(f"Exception when calling EmailCampaignsApi->create_email_campaign: {e}")
            if hasattr(e, 'body'):
                logger.error(f"Sendinblue API response body: {e.body}")
            raise Exception(f'Failed to create Email Campaign with data: {email_campaign_data}') from e
        except Exception as e:
            #this handles any other unexpected exceptions.
            logger.error(f"Unexpected error occurred: {e}")
            raise Exception(f'Unexpected error during Email Campaign creation: {e}') from e



from mindsdb_sql.parser.ast import Describe, Identifier

from mindsdb.api.mongo.classes import Responder
import mindsdb.api.mongo.functions as helpers
from mindsdb.api.mongo.classes.query_sql import run_sql_command


class Responce(Responder):
    when = {'describe': helpers.is_true}

    def result(self, query, request_env, mindsdb_env, session):
        db = query['$db']
        model = query['describe']

        ast_query = Describe(Identifier(model))

        data = run_sql_command(request_env, ast_query)
        res = {'data': data, 'ns': f"{db}.{model}"}

        return res


responder = Responce()



from collections import OrderedDict

from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE


connection_args = OrderedDict(
    alias={
        "type": ARG_TYPE.STR,
        "description": "alias of the Milvus connection to construct",
        "required": True,
    },
    host={
        "type": ARG_TYPE.STR,
        "description": "IP address of the Milvus server",
        "required": True,
    },
    port={
        "type": ARG_TYPE.INT,
        "description": "port of the Milvus server",
        "required": True,
    },
    user={
        "type": ARG_TYPE.STR,
        "description": "username of the Milvus server",
        "required": True,
    },
    password={
        "type": ARG_TYPE.PWD,
        "description": "password of the username of the Milvus server",
        "required": True,
        "secret": True
    },
    search_default_limit={
        "type": ARG_TYPE.INT,
        "description": "default limit to be passed in select statements",
        "required": False,
    },
    search_metric_type={
        "type": ARG_TYPE.STR,
        "description": "metric type used for searches",
        "required": False,
    },
    search_ignore_growing={
        "type": ARG_TYPE.BOOL,
        "description": "whether to ignore growing segments during similarity searches",
        "required": False,
    },
    search_params={
        "type": ARG_TYPE.DICT,
        "description": "specific to the `search_metric_type`",
        "required": False,
    },
    create_auto_id={
        "type": ARG_TYPE.BOOL,
        "description": "whether to auto generate id when inserting records with no ID (default=False)",
        "required": False,
    },
    create_id_max_len={
        "type": ARG_TYPE.STR,
        "description": "maximum length of the id field when creating a table (default=64)",
        "required": False,
    },
    create_embedding_dim={
        "type": ARG_TYPE.INT,
        "description": "embedding dimension for creating table (default=8)",
        "required": False,
    },
    create_dynamic_field={
        "type": ARG_TYPE.BOOL,
        "description": "whether or not the created tables have dynamic fields or not (default=True)",
        "required": False,
    },
    create_content_max_len={
        "type": ARG_TYPE.INT,
        "description": "max length of the content column (default=200)",
        "required": False,
    },
    create_content_default_value={
        "type": ARG_TYPE.STR,
        "description": "default value of content column (default='')",
        "required": False,
    },
    create_schema_description={
        "type": ARG_TYPE.STR,
        "description": "description of the created schemas (default='')",
        "required": False,
    },
    create_alias={
        "type": ARG_TYPE.STR,
        "description": "alias of the created schemas (default='default')",
        "required": False,
    },
    create_index_params={
        "type": ARG_TYPE.DICT,
        "description": "parameters of the index created on embeddings column (default={})",
        "required": False,
    },
    create_index_metric_type={
        "type": ARG_TYPE.STR,
        "description": "metric used to create the index (default='L2')",
        "required": False,
    },
    create_index_type={
        "type": ARG_TYPE.STR,
        "description": "the type of index (default='AUTOINDEX')",
        "required": False,
    },
)

connection_args_example = OrderedDict(
    alias="default",
    host="127.0.0.1",
    port=19530,
    user="username",
    password="password",
    search_default_limit=100,
    search_metric_type="L2",
    search_ignore_growing=True,
    search_params={"nprobe": 10},
    create_auto_id=False,
    create_id_max_len=64,
    create_embedding_dim=8,
    create_dynamic_field=True,
    create_content_max_len=200,
    create_content_default_value="",
    create_schema_description="MindsDB generated table",
    create_alias="default",
    create_index_params={},
    create_index_metric_type="L2",
    create_index_type="AUTOINDEX",
)



import unittest
from mindsdb.api.executor.data_types.response_type import (
    RESPONSE_TYPE,
)
from mindsdb.integrations.handlers.duckdb_handler.duckdb_handler import (
    DuckDBHandler,
)


class DuckDBHandlerTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.kwargs = {'connection_data': {'database': 'db.duckdb'}}
        cls.handler = DuckDBHandler('test_duckdb_handler', **cls.kwargs)

    def test_0_connect(self):
        self.handler.connect()

    def test_1_check_connection(self):
        self.handler.check_connection()

    def test_2_drop_table(self):
        res = self.handler.query('DROP TABLE IF EXISTS integers;')
        assert res.type is not RESPONSE_TYPE.ERROR

    def test_3_create_table(self):
        res = self.handler.query('CREATE TABLE integers(i INTEGER)')
        assert res.type is not RESPONSE_TYPE.ERROR

    def test_4_insert_into_table(self):
        res = self.handler.query('INSERT INTO integers VALUES (42)')
        assert res.type is not RESPONSE_TYPE.ERROR

    def test_5_select(self):
        res = self.handler.query('SELECT * FROM integers;')
        assert res.type is RESPONSE_TYPE.TABLE

    def test_6_describe_table(self):
        res = self.handler.get_columns('integers')
        assert res.type is RESPONSE_TYPE.TABLE

    def test_7_get_tables(self):
        res = self.handler.get_tables()
        assert res.type is not RESPONSE_TYPE.ERROR


if __name__ == '__main__':
    unittest.main()



import spacy
import pandas as pd
from mindsdb.integrations.libs.base import BaseMLEngine
from typing import Optional


def named_entity_recognition(doc):
    pred_whole = set()
    output = {"entity": [], "star_char": [], "end_char": [], "label_": []}
    for ent in doc.ents:
        output["entity"].append(str(ent))
        output["star_char"].append(ent.start_char)
        output["end_char"].append(ent.end_char)
        output["label_"].append(ent.label_)
        pred_whole.add((ent.start_char, ent.end_char, ent.label_))
    return pred_whole, output


def lemmatization(doc):
    pred_whole = set()
    output = {"lemma_": []}
    for token in doc:
        output["lemma_"].append(token.lemma_)
        pred_whole.add((token.lemma_))
    return pred_whole, output


def dependency_parsing(doc):
    pred_whole = set()
    output = {
        "text": [],
        "dep_": [],
        "head.text": [],
        "head.pos_": [],
        "children": [],
    }
    for token in doc:
        children = str([child for child in token.children])
        output["text"].append(token.text)
        output["dep_"].append(token.dep_)
        output["head.text"].append(token.head.text)
        output["head.pos_"].append(token.head.pos_)
        output["children"].append(children)
        pred_whole.add(
            (
                token.text,
                token.dep_,
                token.head.text,
                token.head.pos_,
                children,
            )
        )
    return pred_whole, output


def pos_tagging(doc):
    pred_whole = set()
    output = {
        "text": [],
        "lemma_": [],
        "pos_": [],
        "tag_": [],
        "dep_": [],
        "shape_": [],
        "is_alpha": [],
        "is_stop": [],
    }
    for token in doc:
        output["text"].append(token.text)
        output["lemma_"].append(token.lemma_)
        output["pos_"].append(token.pos_)
        output["tag_"].append(token.tag_)
        output["dep_"].append(token.dep_)
        output["shape_"].append(token.shape_)
        output["is_alpha"].append(token.is_alpha)
        output["is_stop"].append(token.is_stop)
        pred_whole.add(
            (
                token.text,
                token.lemma_,
                token.pos_,
                token.tag_,
                token.dep_,
                token.shape_,
                token.is_alpha,
                token.is_stop,
            )
        )
    return pred_whole, output


def morphology(doc):
    pred_whole = set()
    output = {"token": [], "token.morph": []}
    for token in doc:
        output["token"].append(str(token))
        output["token.morph"].append(str(token.morph))
        pred_whole.add((str(token), str(token.morph)))
    return pred_whole, output


lingustic_features = {
    "ner": named_entity_recognition,
    "lemmatization": lemmatization,
    "dependency-parsing": dependency_parsing,
    "pos-tag": pos_tagging,
    "morphology": morphology,
}


class SpacyHandler(BaseMLEngine):
    """
    Integration with the spaCy NLP library.
    """

    name = "spacy"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.default_linguistic_feature = "ner"
        self.linguistic_features = [
            "ner",
            "lemmatization",
            "dependency-parsing",
            "pos-tag",
            "morphology",
        ]

    @staticmethod
    def create_validation(target, args=None, **kwargs):
        if "using" not in args:
            raise Exception(
                "spaCy engine requires a USING clause! Refer to its documentation for more details."
            )
        else:
            args = args["using"]

        if len(set(args.keys()) & {"linguistic_feature", "target_column"}) == 0:
            raise Exception(
                "`linguistic_feature` and `target_column` are required for this engine."
            )

    def create(
        self,
        target: str,
        df: Optional[pd.DataFrame] = None,
        args: Optional[dict] = None,
    ) -> None:
        model_args = args["using"]
        model_args["target"] = target

        if not args.get("linguistic_feature"):
            args["linguistic_feature"] = self.default_linguistic_feature

        # Loading the model
        nlp = spacy.load("en_core_web_sm")

        # Serialize the model
        bytes_data = nlp.to_bytes()

        # Save the serialized model and its config and model_args to the model storage
        self.model_storage.file_set("model_data", bytes_data)
        self.model_storage.json_set("model_args", model_args)
        self.model_storage.json_set("config", nlp.config)

    def predict(self, df, args=None):
        config = self.model_storage.json_get("config")
        model_args = self.model_storage.json_get("model_args")
        bytes_data = self.model_storage.file_get("model_data")

        # Deserialize the model using the provided config
        lang_cls = spacy.util.get_lang_class(config["nlp"]["lang"])
        nlp = lang_cls.from_config(config)
        nlp.from_bytes(bytes_data)

        column_name = model_args["target_column"]
        linguistic_feature = model_args.get("linguistic_feature")

        predictions_whole = []
        predictions_attributes = []
        for _, text in df.iterrows():
            doc = nlp(text[column_name])
            if linguistic_feature in lingustic_features:
                pred_whole, pred_attributes = lingustic_features[linguistic_feature](
                    doc
                )
                predictions_whole.append(pred_whole)
                predictions_attributes.append(pred_attributes)

        predictions_df = pd.DataFrame(predictions_attributes)
        df = pd.concat([df, predictions_df], axis=1)
        df[model_args["target"]] = predictions_whole

        return df



import re
from typing import Union


def strip_null_byte(x: Union[str, bytes], encoding=None):
    if type(x) == bytes:
        if encoding is None:
            encoding = "UTF-8"
        x = x.decode(encoding=encoding)
    return re.sub(r'[\s\x00]+$', '', x)


from collections import OrderedDict

from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE


connection_args = OrderedDict(
    project_id={
        'type': ARG_TYPE.STR,
        'description': 'The BigQuery project id.'
    },
    dataset={
        'type': ARG_TYPE.STR,
        'description': 'The BigQuery dataset name.'
    },
    service_account_keys={
        'type': ARG_TYPE.PATH,
        'description': 'Full path or URL to the service account JSON file',
        'secret': True
    },
    service_account_json={
        'type': ARG_TYPE.DICT,
        'description': 'Content of service account JSON file',
        'secret': True
    },
)

connection_args_example = OrderedDict(
    project_id='tough-future-332513',
    service_account_keys='/home/bigq/tough-future-332513.json'
)



import json
import pandas as pd
from typing import List

from faunadb import query as q
from faunadb.client import FaunaClient
from mindsdb_sql import Select, Insert, CreateTable, Delete
from mindsdb_sql.parser.ast.select.star import Star
from mindsdb_sql.parser.ast.base import ASTNode

from mindsdb.integrations.libs.response import (
    RESPONSE_TYPE,
    HandlerResponse as Response,
    HandlerStatusResponse as StatusResponse,
)
from mindsdb.integrations.libs.base import DatabaseHandler

from mindsdb.utilities import log

logger = log.getLogger(__name__)


class FaunaDBHandler(DatabaseHandler):
    """This handler handles connection and execution of the FaunaDB statements."""

    name = "faunadb"

    def __init__(self, name: str, **kwargs):
        super().__init__(name)

        self._connection_data = kwargs.get("connection_data")

        self._client_config = {
            "fauna_secret": self._connection_data.get("fauna_secret"),
            "fauna_scheme": self._connection_data.get("fauna_scheme"),
            "fauna_domain": self._connection_data.get("fauna_domain"),
            "fauna_port": self._connection_data.get("fauna_port"),
            "fauna_endpoint": self._connection_data.get("fauna_endpoint"),
        }

        scheme, domain, port, endpoint = (
            self._client_config["fauna_scheme"],
            self._client_config["fauna_domain"],
            self._client_config["fauna_port"],
            self._client_config["fauna_endpoint"],
        )

        # should have the secret
        if (self._client_config["fauna_secret"]) is None:
            raise Exception("FaunaDB secret is required for FaunaDB connection!")
        # either scheme + domain + port or endpoint is required
        # but not both
        if not endpoint and not (scheme and domain and port):
            raise Exception(
                "Either scheme + domain + port or endpoint is required for FaunaDB connection!"
            )
        elif endpoint and (scheme or domain or port):
            raise Exception(
                "Either scheme + domain + port or endpoint is required for FaunaDB connection, but not both!"
            )

        self._client = None
        self.is_connected = False
        self.connect()

    def _get_client(self):
        client_config = self._client_config
        if client_config is None:
            raise Exception("Client config is not set!")

        if client_config["fauna_endpoint"] is not None:
            return FaunaClient(
                secret=client_config["fauna_secret"],
                endpoint=client_config["fauna_endpoint"],
            )
        else:
            return FaunaClient(
                secret=client_config["fauna_secret"],
                scheme=client_config["fauna_scheme"],
                domain=client_config["fauna_domain"],
                port=client_config["fauna_port"],
            )

    def __del__(self):
        if self.is_connected is True:
            self.disconnect()

    def connect(self):
        """Connect to a FaunaDB database."""
        if self.is_connected is True:
            return self._client

        try:
            self._client = self._get_client()
            self.is_connected = True
            return self._client
        except Exception as e:
            logger.error(f"Error connecting to FaunaDB client, {e}!")
            self.is_connected = False

    def disconnect(self):
        """Close the database connection."""

        if self.is_connected is False:
            return

        self._client = None
        self.is_connected = False

    def check_connection(self):
        """Check the connection to the FaunaDB database."""
        response_code = StatusResponse(False)
        need_to_close = self.is_connected is False

        try:
            self._client.ping()
            response_code.success = True
        except Exception as e:
            logger.error(f"Error connecting to FaunaDB , {e}!")
            response_code.error_message = str(e)
        finally:
            if response_code.success is True and need_to_close:
                self.disconnect()
            if response_code.success is False and self.is_connected is True:
                self.is_connected = False

        return response_code

    def query(self, query: ASTNode) -> Response:
        """Render and execute a SQL query.

        Args:
            query (ASTNode): The SQL query.

        Returns:
            Response: The query result.
        """

        if isinstance(query, Select):
            collection = str(query.from_table)
            fields = query.targets
            conditions = query.where
            offset = query.offset
            limit = query.limit
            # TODO: research how to parse individual columns from document
            # fields = [f.to_string().split()[2] for f in fields]
            result = self.select(collection, fields, conditions, offset, limit)

        elif isinstance(query, Insert):
            collection = str(query.table)
            fields = [col.name for col in query.columns]
            values = query.values
            self.insert(collection, fields, values)
            return Response(resp_type=RESPONSE_TYPE.OK)

        elif isinstance(query, CreateTable):
            collection_name = str(query.name)
            self.create_table(collection_name)
            return Response(resp_type=RESPONSE_TYPE.OK)

        elif isinstance(query, Delete):
            collection_name = str(query.table)
            conditions = query.where
            self.delete_document(collection_name, conditions)
            return Response(resp_type=RESPONSE_TYPE.OK)
        """
        # NOT Working for integration tables yet
        elif isinstance(query, DropTables):
            collection_name = str(query.tables)
            self.drop_table(collection_name)
        """

        df = pd.json_normalize(result['data'])
        return Response(RESPONSE_TYPE.TABLE, df)

    def select(
        self,
        table_name: str,
        columns: List[str] = None,
        conditions: List[str] = None,
        offset: int = None,
        limit: int = None,
    ) -> dict:
        # select * from db_name.collection_name
        if len(columns) > 0 and isinstance(columns[0], Star):
            fauna_query = q.map_(
                q.lambda_("ref", q.get(q.var("ref"))),
                q.paginate(q.documents(q.collection(str(table_name)))),
            )
        else:
            # select id, name ,etc from db_name.collection_name
            fauna_query = q.map_(
                q.lambda_("doc", {"data": q.select(columns, q.var("doc"))}),
                q.paginate(q.documents(q.collection(table_name))),
            )

        return self._client.query(fauna_query)

    def insert(self, table_name: str, fields, values) -> Response:
        if len(fields) == 1 and fields[0] == "data":
            for value in values:
                value = json.loads(value[0])
                if isinstance(value, dict):
                    value = [value]
                for data in value:
                    self._client.query(
                        q.create(
                            q.collection(table_name),
                            {"data": data},
                        )
                    )
        else:
            for value in values:
                data = {f: v for f, v in zip(fields, value)}
                self._client.query(
                    q.create(
                        q.collection(table_name),
                        {"data": data},
                    )
                )

    def create_table(self, table_name: str, if_not_exists=True) -> Response:
        """
        Create a collection with the given name in the FaunaDB database.
        """
        fauna_query = q.create_collection({"name": table_name})
        self._client.query(fauna_query)

    def delete_document(self, table_name: str, conditions: List[str]) -> Response:
        """
        Delete a document with the given id in the FaunaDB database.
        """
        # get the id of the document (only = operator supported right now, can add more)
        ref = conditions.args[1].value
        fauna_query = q.delete(q.ref(q.collection(table_name), ref))
        self._client.query(fauna_query)

    def drop_table(self, table_name: str, if_exists=True) -> Response:
        """
        Delete a collection from the FaunaDB database.
        """
        fauna_query = q.delete(q.collection(table_name))
        self._client.query(fauna_query)
        return Response(resp_type=RESPONSE_TYPE.OK)

    def get_tables(self) -> Response:
        """
        Get the list of collections in the FaunaDB database.
        """
        try:
            result = self._client.query(q.paginate(q.collections()))
            collections = []
            for collection in result["data"]:
                collections.append(collection.id())
            return Response(
                resp_type=RESPONSE_TYPE.TABLE,
                data_frame=pd.DataFrame(
                    collections,
                    columns=["table_name"],
                ),
            )
        except Exception as e:
            logger.error(f"Error getting tables from FaunaDB: {e}")
            return Response(
                resp_type=RESPONSE_TYPE.ERROR,
                error_message=f"Error getting tables from FaunaDB: {e}",
            )



from mindsdb_sql import parse_sql

from mindsdb.integrations.handlers.luma_handler.luma_tables import (
    LumaEventsTable
)
from mindsdb.integrations.handlers.luma_handler.luma import LumaClient
from mindsdb.integrations.libs.api_handler import APIHandler
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
)


class LumaHandler(APIHandler):
    """The Luma handler implementation"""

    def __init__(self, name: str, **kwargs):
        """Initialize the Luma handler.

        Parameters
        ----------
        name : str
            name of a handler instance
        """
        super().__init__(name)

        connection_data = kwargs.get("connection_data", {})
        self.connection_data = connection_data
        self.kwargs = kwargs
        self.luma_client = None
        self.is_connected = False

        events_data = LumaEventsTable(self)
        self._register_table("events", events_data)

    def connect(self) -> StatusResponse:
        """Set up the connection required by the handler.

        Returns
        -------
        StatusResponse
            connection object
        """
        resp = StatusResponse(False)
        try:
            self.luma_client = LumaClient(self.connection_data.get("api_key"))
            resp.success = True
            self.is_connected = True
        except Exception as ex:
            resp.success = False
            resp.error_message = ex
        return resp

    def check_connection(self) -> StatusResponse:
        """Check connection to the handler.

        Returns
        -------
        StatusResponse
            Status confirmation
        """
        return self.connect()

    def native_query(self, query: str) -> StatusResponse:
        """Receive and process a raw query.

        Parameters
        ----------
        query : str
            query in a native format

        Returns
        -------
        StatusResponse
            Request status
        """
        ast = parse_sql(query, dialect="mindsdb")
        return self.query(ast)



from http import HTTPStatus
from typing import List

from flask import request
from flask_restx import Resource

import pandas as pd

from mindsdb.api.http.namespaces.configs.projects import ns_conf
from mindsdb.api.executor.controllers.session_controller import SessionController
from mindsdb.api.http.utils import http_error
from mindsdb.metrics.metrics import api_endpoint_metrics
from mindsdb.integrations.handlers.web_handler.urlcrawl_helpers import get_all_websites
from mindsdb.interfaces.database.projects import ProjectController
from mindsdb.interfaces.file.file_controller import FileController
from mindsdb.integrations.utilities.rag.loaders.file_loader import FileLoader
from mindsdb.integrations.utilities.rag.splitters.file_splitter import FileSplitter, FileSplitterConfig
from mindsdb.interfaces.knowledge_base.controller import KnowledgeBaseTable
from mindsdb.utilities import log

logger = log.getLogger(__name__)

_DEFAULT_MARKDOWN_HEADERS_TO_SPLIT_ON = [
    ("#", "Header 1"),
    ("##", "Header 2"),
    ("###", "Header 3"),
]


def _insert_file_into_knowledge_base(table: KnowledgeBaseTable, file_name: str):
    file_controller = FileController()
    splitter = FileSplitter(FileSplitterConfig())
    file_path = file_controller.get_file_path(file_name)
    loader = FileLoader(file_path)
    split_docs = []
    for doc in loader.lazy_load():
        split_docs += splitter.split_documents([doc])
    doc_objs = []
    for split_doc in split_docs:
        doc_objs.append({
            'content': split_doc.page_content,
        })
    docs_df = pd.DataFrame.from_records(doc_objs)
    # Insert documents into KB
    table.insert(docs_df)


def _insert_web_pages_into_knowledge_base(table: KnowledgeBaseTable, urls: List[str]):
    try:
        # To prevent dependency on langchain_text_splitters unless needed.
        from langchain_text_splitters import MarkdownHeaderTextSplitter
    except ImportError as e:
        logger.error(f'Error importing langchain_text_splitters to insert web page into knowledge base: {e}')
        raise e

    websites_df = get_all_websites(urls)
    # Text content is treated as markdown.
    markdown_splitter = MarkdownHeaderTextSplitter(headers_to_split_on=_DEFAULT_MARKDOWN_HEADERS_TO_SPLIT_ON)

    def append_row_documents(row, all_docs):
        docs = markdown_splitter.split_text(row['text_content'])
        for doc in docs:
            # Link the URL to each web page chunk.
            doc.metadata['url'] = row['url']
        all_docs += docs

    all_docs = []
    websites_df.apply(lambda row: append_row_documents(row, all_docs), axis=1)
    # Convert back to a DF.
    doc_objs = []
    for doc in all_docs:
        doc_objs.append({
            'content': doc.page_content,
            'url': doc.metadata['url']
        })
    docs_df = pd.DataFrame.from_records(doc_objs)
    # Insert documents into KB.
    table.insert(docs_df)


@ns_conf.route('/<project_name>/knowledge_bases/<knowledge_base_name>')
@ns_conf.param('project_name', 'Name of the project')
@ns_conf.param('knowledge_base_name', 'Name of the knowledge_base')
class KnowledgeBaseResource(Resource):
    @ns_conf.doc('update_knowledge_base')
    @api_endpoint_metrics('PUT', '/knowledge_bases/knowledge_base')
    def put(self, project_name: str, knowledge_base_name: str):
        '''Updates a knowledge base.'''
        # Check for required parameters.
        if 'knowledge_base' not in request.json:
            return http_error(
                HTTPStatus.BAD_REQUEST,
                'Missing parameter',
                'Must provide "knowledge_base" parameter in PUT body'
            )
        session = SessionController()
        project_controller = ProjectController()
        try:
            project = project_controller.get(name=project_name)
        except ValueError:
            # Project must exist.
            return http_error(
                HTTPStatus.NOT_FOUND,
                'Project not found',
                f'Project with name {project_name} does not exist'
            )
        try:
            table = session.kb_controller.get_table(knowledge_base_name, project.id)
        except ValueError:
            # Knowledge Base must exist.
            return http_error(
                HTTPStatus.NOT_FOUND,
                'Knowledge Base not found',
                f'Knowledge Base with name {knowledge_base_name} does not exist'
            )

        kb = request.json['knowledge_base']
        files = kb.get('files', [])
        urls = kb.get('urls', [])

        # Load, split, & embed files into Knowledge Base.
        for file_name in files:
            _insert_file_into_knowledge_base(table, file_name)
        # Crawl, split, & embed web pages into Knowledge Base.
        _insert_web_pages_into_knowledge_base(table, urls)
        return '', HTTPStatus.OK



from typing import Dict, Optional

import pandas as pd
import transformers
from huggingface_hub import HfApi

from mindsdb.integrations.handlers.huggingface_handler.settings import FINETUNE_MAP
from mindsdb.integrations.libs.base import BaseMLEngine
from mindsdb.utilities import log

logger = log.getLogger(__name__)


class HuggingFaceHandler(BaseMLEngine):
    name = "huggingface"

    @staticmethod
    def create_validation(target, args=None, **kwargs):

        if "using" in args:
            args = args["using"]

        hf_api = HfApi()

        # check model is pytorch based
        metadata = hf_api.model_info(args["model_name"])
        if "pytorch" not in metadata.tags:
            raise Exception(
                "Currently only PyTorch models are supported (https://huggingface.co/models?library=pytorch&sort=downloads). To request another library, please contact us on our community slack (https://mindsdb.com/joincommunity)."
            )

        # check model task
        supported_tasks = [
            "text-classification",
            "text-generation",
            "zero-shot-classification",
            "translation",
            "summarization",
            "text2text-generation",
            "fill-mask",
        ]

        if metadata.pipeline_tag not in supported_tasks:
            raise Exception(
                f'Not supported task for model: {metadata.pipeline_tag}.\
             Should be one of {", ".join(supported_tasks)}'
            )

        if "task" not in args:
            args["task"] = metadata.pipeline_tag
        elif args["task"] != metadata.pipeline_tag:
            raise Exception(
                f'Task mismatch for model: {args["task"]}!={metadata.pipeline_tag}'
            )

        input_keys = list(args.keys())

        # task, model_name, input_column is essential
        for key in ["task", "model_name", "input_column"]:
            if key not in args:
                raise Exception(f'Parameter "{key}" is required')
            input_keys.remove(key)

        # check tasks input

        if args["task"] == "zero-shot-classification":
            key = "candidate_labels"
            if key not in args:
                raise Exception(
                    '"candidate_labels" is required for zero-shot-classification'
                )
            input_keys.remove(key)

        if args["task"] == "translation":
            keys = ["lang_input", "lang_output"]
            for key in keys:
                if key not in args:
                    raise Exception(f"{key} is required for translation")
                input_keys.remove(key)

        if args["task"] == "summarization":
            keys = ["min_output_length", "max_output_length"]
            for key in keys:
                if key not in args:
                    raise Exception(f"{key} is required for summarization")
                input_keys.remove(key)

        # optional keys
        for key in ["labels", "max_length", "truncation_policy"]:
            if key in input_keys:
                input_keys.remove(key)

        if len(input_keys) > 0:
            raise Exception(f'Not expected parameters: {", ".join(input_keys)}')

    def create(self, target, args=None, **kwargs):
        # TODO change BaseMLEngine api?
        if "using" in args:
            args = args["using"]

        args["target"] = target

        model_name = args["model_name"]
        hf_model_storage_path = self.engine_storage.folder_get(model_name)  # real

        if args["task"] == "translation":
            args[
                "task_proper"
            ] = f"translation_{args['lang_input']}_to_{args['lang_output']}"
        else:
            args["task_proper"] = args["task"]

        logger.debug(f"Checking file system for {model_name}...")

        ####
        # Check if pipeline has already been downloaded
        try:
            pipeline = transformers.pipeline(task=args['task_proper'], model=hf_model_storage_path,
                                             tokenizer=hf_model_storage_path)
            logger.debug(f'Model already downloaded!')
        ####
        # Otherwise download it
        except OSError:
            try:
                logger.debug(f"Downloading {model_name}...")
                pipeline = transformers.pipeline(task=args['task_proper'], model=model_name)

                pipeline.save_pretrained(hf_model_storage_path)

                logger.debug(f"Saved to {hf_model_storage_path}")
            except Exception:
                raise Exception(
                    "Error while downloading and setting up the model. Please try a different model. We're working on expanding the list of supported models, so we would appreciate it if you let us know about this in our community slack (https://mindsdb.com/joincommunity)."
                )  # noqa
        ####

        if "max_length" in args:
            pass
        elif "max_position_embeddings" in pipeline.model.config.to_dict().keys():
            args["max_length"] = pipeline.model.config.max_position_embeddings
        elif "max_length" in pipeline.model.config.to_dict().keys():
            args["max_length"] = pipeline.model.config.max_length
        else:
            logger.debug('No max_length found!')

        labels_default = pipeline.model.config.id2label
        labels_map = {}
        if "labels" in args:
            for num in labels_default.keys():
                labels_map[labels_default[num]] = args["labels"][num]
            args["labels_map"] = labels_map
        else:
            for num in labels_default.keys():
                labels_map[labels_default[num]] = labels_default[num]
            args["labels_map"] = labels_map

        ###### store and persist in model folder
        self.model_storage.json_set("args", args)

        ###### persist changes to handler folder
        self.engine_storage.folder_sync(model_name)

    # todo move infer tasks to a seperate file
    def predict_text_classification(self, pipeline, item, args):
        top_k = args.get("top_k", 1000)

        result = pipeline(
            [item], top_k=top_k, truncation=True, max_length=args["max_length"]
        )[0]

        final = {}
        explain = {}
        if type(result) == dict:
            result = [result]
        final[args["target"]] = args["labels_map"][result[0]["label"]]
        for elem in result:
            if args["labels_map"]:
                explain[args["labels_map"][elem["label"]]] = elem["score"]
            else:
                explain[elem["label"]] = elem["score"]
        final[f"{args['target']}_explain"] = explain
        return final

    def predict_text_generation(self, pipeline, item, args):
        result = pipeline([item], max_length=args["max_length"])[0]

        final = {}
        final[args["target"]] = result["generated_text"]

        return final

    def predict_zero_shot(self, pipeline, item, args):
        top_k = args.get("top_k", 1000)

        result = pipeline(
            [item],
            candidate_labels=args["candidate_labels"],
            truncation=True,
            top_k=top_k,
            max_length=args["max_length"],
        )[0]

        final = {}
        final[args["target"]] = result["labels"][0]

        explain = dict(zip(result["labels"], result["scores"]))
        final[f"{args['target']}_explain"] = explain

        return final

    def predict_translation(self, pipeline, item, args):
        result = pipeline([item], max_length=args["max_length"])[0]

        final = {}
        final[args["target"]] = result["translation_text"]

        return final

    def predict_summarization(self, pipeline, item, args):
        result = pipeline(
            [item],
            min_length=args["min_output_length"],
            max_length=args["max_output_length"],
        )[0]

        final = {}
        final[args["target"]] = result["summary_text"]

        return final

    def predict_text2text(self, pipeline, item, args):
        result = pipeline([item], max_length=args["max_length"])[0]

        final = {}
        final[args["target"]] = result["generated_text"]

        return final

    def predict_fill_mask(self, pipeline, item, args):
        result = pipeline([item])[0]

        final = {}
        final[args["target"]] = result[0]["sequence"]
        explain = {elem["sequence"]: elem["score"] for elem in result}
        final[f"{args['target']}_explain"] = explain

        return final

    def predict(self, df, args=None):

        fnc_list = {
            "text-classification": self.predict_text_classification,
            "text-generation": self.predict_text_generation,
            "zero-shot-classification": self.predict_zero_shot,
            "translation": self.predict_translation,
            "summarization": self.predict_summarization,
            "fill-mask": self.predict_fill_mask,
        }

        ###### get stuff from model folder
        args = self.model_storage.json_get("args")

        task = args["task"]

        if task not in fnc_list:
            raise RuntimeError(f"Unknown task: {task}")

        fnc = fnc_list[task]

        try:
            # load from model storage (finetuned models will use this)
            hf_model_storage_path = self.model_storage.folder_get(
                args["model_name"]
            )
            pipeline = transformers.pipeline(
                task=args["task_proper"],
                model=hf_model_storage_path,
                tokenizer=hf_model_storage_path,
            )
        except OSError:
            # load from engine storage (i.e. 'common' models)
            hf_model_storage_path = self.engine_storage.folder_get(
                args["model_name"]
            )
            pipeline = transformers.pipeline(
                task=args["task_proper"],
                model=hf_model_storage_path,
                tokenizer=hf_model_storage_path,
            )

        input_column = args["input_column"]
        if input_column not in df.columns:
            raise RuntimeError(f'Column "{input_column}" not found in input data')
        input_list = df[input_column]

        max_tokens = pipeline.tokenizer.model_max_length

        results = []
        for item in input_list:
            if max_tokens is not None:
                tokens = pipeline.tokenizer.encode(item)
                if len(tokens) > max_tokens:
                    truncation_policy = args.get("truncation_policy", "strict")
                    if truncation_policy == "strict":
                        results.append(
                            {
                                "error": f"Tokens count exceed model limit: {len(tokens)} > {max_tokens}"
                            }
                        )
                        continue
                    elif truncation_policy == "left":
                        tokens = tokens[
                            -max_tokens + 1 : -1
                        ]  # cut 2 empty tokens from left and right
                    else:
                        tokens = tokens[
                            1 : max_tokens - 1
                        ]  # cut 2 empty tokens from left and right

                    item = pipeline.tokenizer.decode(tokens)

            item = str(item)
            try:
                result = fnc(pipeline, item, args)
            except Exception as e:
                msg = str(e).strip()
                if msg == "":
                    msg = e.__class__.__name__
                result = {"error": msg}
            results.append(result)

        pred_df = pd.DataFrame(results)

        return pred_df

    def describe(self, attribute: Optional[str] = None) -> pd.DataFrame:
        args = self.model_storage.json_get("args")
        if attribute == "args":
            return pd.DataFrame(args.items(), columns=["key", "value"])
        elif attribute == "metadata":
            hf_api = HfApi()
            metadata = hf_api.model_info(args["model_name"])
            data = metadata.__dict__
            return pd.DataFrame(list(data.items()), columns=["key", "value"])
        else:
            tables = ["args", "metadata"]
            return pd.DataFrame(tables, columns=["tables"])

    def finetune(
        self, df: Optional[pd.DataFrame] = None, args: Optional[Dict] = None
    ) -> None:
        finetune_args = args if args else {}
        args = self.base_model_storage.json_get("args")
        args.update(finetune_args)

        model_name = args["model_name"]
        model_folder = self.model_storage.folder_get(model_name)
        args["model_folder"] = model_folder
        model_folder_name = model_folder.split("/")[-1]
        task = args["task"]

        if task not in FINETUNE_MAP:
            raise KeyError(
                f"{task} is not currently supported, please choose a supported task - {', '.join(FINETUNE_MAP)}"
            )

        tokenizer, trainer = FINETUNE_MAP[task](df, args)

        try:
            trainer.train()
            trainer.save_model(
                model_folder
            )  # TODO: save entire pipeline instead https://huggingface.co/docs/transformers/main_classes/pipelines#transformers.Pipeline.save_pretrained
            tokenizer.save_pretrained(model_folder)

            # persist changes
            self.model_storage.json_set("args", args)
            self.model_storage.folder_sync(model_folder_name)

        except Exception as e:
            err_str = f"Finetune failed with error: {str(e)}"
            logger.debug(err_str)
            raise Exception(err_str)



from mindsdb.integrations.handlers.scylla_handler import Handler as ScyllaHandler


class CassandraHandler(ScyllaHandler):
    """
    This handler handles connection and execution of the Cassandra statements.
    """

    name = 'cassandra'

    def __init__(self, name, **kwargs):
        super().__init__(name, **kwargs)



import copy
from typing import List

import pandas as pd

import mindsdb_sql.planner.utils as utils
from mindsdb_sql.parser.ast import (
    BinaryOperation,
    Constant,
    Identifier,
    Select,
    Update,
    Delete,
    Star
)
from mindsdb_sql.parser.dialects.mindsdb import CreatePredictor

import mindsdb.interfaces.storage.db as db
from mindsdb.integrations.libs.vectordatabase_handler import TableField
from mindsdb.interfaces.database.projects import ProjectController
from mindsdb.utilities.exception import EntityExistsError, EntityNotExistsError


class KnowledgeBaseTable:
    """
    Knowledge base table interface
    Handlers requests to KB table and modifies data in linked vector db table
    """

    def __init__(self, kb: db.KnowledgeBase, session):
        self._kb = kb
        self._vector_db = None
        self.session = session

    def select_query(self, query: Select) -> pd.DataFrame:
        """
        Handles select from KB table.
        Replaces content values with embeddings in where clause. Sends query to vector db
        :param query: query to KB table
        :return: dataframe with the result table
        """

        # replace content with embeddings

        utils.query_traversal(query.where, self._replace_query_content)

        # set table name
        query.from_table = Identifier(parts=[self._kb.vector_database_table])

        # remove embeddings from result
        targets = []
        for target in query.targets:
            if isinstance(target, Star):
                targets.extend([
                    Identifier(TableField.ID.value),
                    Identifier(TableField.CONTENT.value),
                    Identifier(TableField.METADATA.value),
                ])
            elif isinstance(target, Identifier) and target.parts[-1].lower() != TableField.EMBEDDINGS.value:
                targets.append(target)
        query.targets = targets

        # send to vectordb
        db_handler = self._get_vector_db()
        resp = db_handler.query(query)
        return resp.data_frame

    def update_query(self, query: Update):
        """
        Handles update query to KB table.
        Replaces content values with embeddings in SET clause. Sends query to vector db
        :param query: query to KB table
        """

        # add embeddings to content in updated collumns
        query = copy.deepcopy(query)

        emb_col = TableField.EMBEDDINGS.value
        cont_col = TableField.CONTENT.value
        if cont_col in query.update_columns:
            content = query.update_columns[cont_col]
            query.update_columns[emb_col] = Constant(self._content_to_embeddings(content))

        # TODO search content in where clause?

        # set table name
        query.table = Identifier(parts=[self._kb.vector_database_table])

        # send to vectordb
        db_handler = self._get_vector_db()
        db_handler.query(query)

    def delete_query(self, query: Delete):
        """
        Handles delete query to KB table.
        Replaces content values with embeddings in WHERE clause. Sends query to vector db
        :param query: query to KB table
        """
        utils.query_traversal(query.where, self._replace_query_content)

        # set table name
        query.table = Identifier(parts=[self._kb.vector_database_table])

        # send to vectordb
        db_handler = self._get_vector_db()
        db_handler.query(query)

    def clear(self):
        """
        Clear data in KB table
        Sends delete to vector db table
        """
        db_handler = self._get_vector_db()
        db_handler.delete(self._kb.vector_database_table)

    def insert(self, df: pd.DataFrame):
        """
        Insert dataframe to KB table
        Adds embedding column to dataframe and calls .upsert method of vector db
        :param df: input dataframe

        """
        if df.empty:
            return

        df = self._adapt_column_names(df)

        # add embeddings
        df_emb = self._df_to_embeddings(df)
        df = pd.concat([df, df_emb], axis=1)

        # send to vector db
        db_handler = self._get_vector_db()
        db_handler.do_upsert(self._kb.vector_database_table, df)

    def _adapt_column_names(self, df: pd.DataFrame) -> pd.DataFrame:

        '''
            convert input columns for vector db input
            - id, content and metadata
        '''

        params = self._kb.params

        columns = list(df.columns)

        # -- prepare id --

        # if id_column is defined:
        #     use it as id
        # elif 'id' column exists:
        #     use it
        # else:
        #     use hash(content) -- it happens inside of vector handler

        id_column = params.get('id_column')
        if id_column is not None and id_column not in columns:
            # wrong name
            id_column = None

        if id_column is None and TableField.ID.value in columns:
            # default value
            id_column = TableField.ID.value

        if id_column is not None:
            # remove from lookup list
            columns.remove(id_column)

        # -- prepare content and metadata --

        # if content_columns is defined:
        #     if len(content_columns) > 1:
        #          make text from row (col: value\n col: value)
        #     if metadata_columns is defined:
        #          use them as metadata
        #     else:
        #          use all unused columns is metadata
        #     elif metadata_columns is defined:
        #          metadata_columns go to metadata
        #          use all unused columns  as content (make text if columns>1)
        # else:
        #     no metadata
        #     all unused columns go to content (make text if columns>1)

        content_columns = params.get('content_columns')
        metadata_columns = params.get('metadata_columns')

        if content_columns is not None:
            content_columns = list(set(content_columns).intersection(columns))
            if len(content_columns) == 0:
                raise ValueError(f'Content columns {params.get("content_columns")} not found in dataset: {columns}')

            if metadata_columns is not None:
                metadata_columns = list(set(metadata_columns).intersection(columns))
            else:
                # all the rest columns
                metadata_columns = list(set(columns).difference(content_columns))

        elif metadata_columns is not None:
            metadata_columns = list(set(metadata_columns).intersection(columns))
            # use all unused columns is content
            content_columns = list(set(columns).difference(metadata_columns))
        else:
            # all columns go to content
            content_columns = columns

        if not content_columns:
            raise ValueError("Can't find content columns")

        def row_to_document(row: pd.Series) -> str:
            """
            Convert a row in the input dataframe into a document

            Default implementation is to concatenate all the columns
            in the form of
            field1: value1\nfield2: value2\n...
            """
            fields = row.index.tolist()
            values = row.values.tolist()
            document = "\n".join(
                [f"{field}: {value}" for field, value in zip(fields, values)]
            )
            return document

        # create dataframe
        if len(content_columns) == 1:
            c_content = df[content_columns[0]]
        else:
            c_content = df[content_columns].apply(row_to_document, axis=1)
        c_content.name = TableField.CONTENT.value
        df_out = pd.DataFrame(c_content)

        if id_column is not None:
            df_out[TableField.ID.value] = df[id_column]

        if metadata_columns and len(metadata_columns) > 0:
            df_out[TableField.METADATA.value] = df[metadata_columns].apply(lambda row: str(dict(row)), axis=1)

        return df_out

    def _replace_query_content(self, node, **kwargs):
        if isinstance(node, BinaryOperation):
            if isinstance(node.args[0], Identifier) and isinstance(node.args[1], Constant):
                col_name = node.args[0].parts[-1]
                if col_name.lower() == TableField.CONTENT.value:
                    # replace
                    node.args[0].parts = [TableField.EMBEDDINGS.value]
                    node.args[1].value = [self._content_to_embeddings(node.args[1].value)]

    def _get_vector_db(self):
        """
        helper to get vector db handler
        """
        if self._vector_db is None:
            database = db.Integration.query.get(self._kb.vector_database_id)
            if database is None:
                raise ValueError('Vector database not found. Is it deleted?')
            database_name = database.name
            self._vector_db = self.session.integration_controller.get_data_handler(database_name)
        return self._vector_db

    def _df_to_embeddings(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        Returns embeddings for input dataframe.
        Uses model embedding model to convert content to embeddings.
        Automatically detects input and output of model using model description
        :param df:
        :return: dataframe with embeddings
        """

        if df.empty:
            return pd.DataFrame([], columns=[TableField.EMBEDDINGS.value])

        model_id = self._kb.embedding_model_id
        # get the input columns
        model_rec = db.session.query(db.Predictor).filter_by(id=model_id).first()

        assert model_rec is not None, f"Model not found: {model_id}"
        model_project = db.session.query(db.Project).filter_by(id=model_rec.project_id).first()

        project_datanode = self.session.datahub.get(model_project.name)

        # keep only content
        df = df[[TableField.CONTENT.value]]

        input_col = model_rec.learn_args.get('using', {}).get('question_column')

        if input_col is not None and input_col != TableField.CONTENT.value:
            df = df.rename(columns={TableField.CONTENT.value: input_col})

        data = df.to_dict('records')

        df_out = project_datanode.predict(
            model_name=model_rec.name,
            data=data,
        )

        target = model_rec.to_predict[0]
        if target != TableField.EMBEDDINGS.value:
            # adapt output for vectordb
            df_out = df_out.rename(columns={target: TableField.EMBEDDINGS.value})
        df_out = df_out[[TableField.EMBEDDINGS.value]]

        return df_out

    def _content_to_embeddings(self, content: str) -> List[float]:
        """
        Converts string to embeddings
        :param content: input string
        :return: embeddings
        """
        df = pd.DataFrame([[content]], columns=[TableField.CONTENT.value])
        res = self._df_to_embeddings(df)
        return res[TableField.EMBEDDINGS.value][0]


class KnowledgeBaseController:
    """
    Knowledge base controller handles all
    manages knowledge bases
    """

    def __init__(self, session) -> None:
        self.session = session

    def add(
        self,
        name: str,
        project_name: str,
        embedding_model: Identifier,
        storage: Identifier,
        params: dict,
        if_not_exists: bool = False,
    ) -> db.KnowledgeBase:
        """
        Add a new knowledge base to the database
        """
        # check if knowledge base already exists

        # get project id

        project = self.session.database_controller.get_project(project_name)

        project_id = project.id

        # not difference between cases in sql
        name = name.lower()

        kb = self.get(name, project_id)
        if kb is not None:
            if if_not_exists:
                return kb
            raise EntityExistsError("Knowledge base already exists", name)

        if embedding_model is None:
            # create default embedding model
            model_name = self._create_default_embedding_model(project.name, name, params=params)

            # memorize to remove it later
            params['embedding_model'] = model_name

        else:
            # get embedding model from input
            model_name = embedding_model.parts[-1]

        if embedding_model is not None and len(embedding_model.parts) > 1:
            # model project is set
            model_project = self.session.database_controller.get_project(embedding_model.parts[-2])
        else:
            model_project = project

        model = self.session.model_controller.get_model(
            name=model_name,
            project_name=model_project.name
        )
        model_record = db.Predictor.query.get(model['id'])
        embedding_model_id = model_record.id

        # search for the vector database table
        if storage is None:
            # create chroma db with same name
            vector_table_name = "default_collection"
            vector_db_name = self._create_persistent_chroma(
                name
            )

            # memorize to remove it later
            params['vector_storage'] = vector_db_name
        elif len(storage.parts) != 2:
            raise ValueError('Storage param has to be vector db with table')
        else:
            vector_db_name, vector_table_name = storage.parts

        vector_database_id = self.session.integration_controller.get(vector_db_name)['id']

        # create table in vectordb
        self.session.datahub.get(vector_db_name).integration_handler.create_table(
            vector_table_name
        )

        kb = db.KnowledgeBase(
            name=name,
            project_id=project_id,
            vector_database_id=vector_database_id,
            vector_database_table=vector_table_name,
            embedding_model_id=embedding_model_id,
            params=params,
        )
        db.session.add(kb)
        db.session.commit()
        return kb

    def _create_persistent_chroma(self, kb_name, engine="chromadb"):
        """Create default vector database for knowledge base, if not specified"""

        vector_store_name = f"{kb_name}_{engine}"

        vector_store_folder_name = f"{vector_store_name}"
        connection_args = {"persist_directory": vector_store_folder_name}

        # check if exists
        if self.session.integration_controller.get(vector_store_name):
            return vector_store_name

        self.session.integration_controller.add(vector_store_name, engine, connection_args)
        return vector_store_name

    def _create_default_embedding_model(self, project_name, kb_name, engine="langchain_embedding", params: dict = None):
        """create a default embedding model for knowledge base, if not specified"""
        model_name = f"{kb_name}_default_model"
        using_args = {}
        if engine == 'langchain_embedding':
            # Use default embeddings.
            using_args['class'] = 'openai'
        # Include API key if provided.
        using_args.update({k: v for k, v in params.items() if 'api_key' in k})
        statement = CreatePredictor(
            name=Identifier(parts=[project_name, model_name]),
            using=using_args,
            targets=[
                Identifier(parts=[TableField.EMBEDDINGS.value])
            ]
        )
        ml_handler = self.session.integration_controller.get_ml_handler(engine)

        _ = self.session.model_controller.create_model(
            statement,
            ml_handler
        )

        return model_name

    def delete(self, name: str, project_name: int, if_exists: bool = False) -> None:
        """
        Delete a knowledge base from the database
        """
        try:
            project = self.session.database_controller.get_project(project_name)
        except ValueError:
            raise ValueError(f"Project not found: {project_name}")
        project_id = project.id

        # check if knowledge base exists
        kb = self.get(name, project_id)
        if kb is None:
            # knowledge base does not exist
            if if_exists:
                return
            else:
                raise EntityNotExistsError("Knowledge base does not exist", name)

        # drop table
        vector_db = db.Integration.query.get(kb.vector_database_id)
        if vector_db:
            database_name = vector_db.name
            self.session.datahub.get(database_name).integration_handler.drop_table(
                kb.vector_database_table
            )

        # drop objects if they were created automatically
        if 'vector_storage' in kb.params:
            try:
                self.session.integration_controller.delete(kb.params['vector_storage'])
            except EntityNotExistsError:
                pass
        if 'embedding_model' in kb.params:
            try:
                self.session.model_controller.delete_model(kb.params['embedding_model'], project_name)
            except EntityNotExistsError:
                pass

        # kb exists
        db.session.delete(kb)
        db.session.commit()

    def get(self, name: str, project_id: int) -> db.KnowledgeBase:
        """
        Get a knowledge base from the database
        by name + project_id
        """
        kb = (
            db.session.query(db.KnowledgeBase)
            .filter_by(
                name=name,
                project_id=project_id,
            )
            .first()
        )
        return kb

    def get_table(self, name: str, project_id: int) -> KnowledgeBaseTable:
        """
        Returns kb table object
        :param name: table name
        :param project_id: project id
        :return: kb table object
        """
        kb = self.get(name, project_id)
        if kb is not None:
            return KnowledgeBaseTable(kb, self.session)

    def list(self, project_name: str = None) -> List[dict]:
        """
        List all knowledge bases from the database
        belonging to a project
        """
        project_controller = ProjectController()
        projects = project_controller.get_list()
        if project_name is not None:
            projects = [p for p in projects if p.name == project_name]

        query = (
            db.session.query(db.KnowledgeBase)
            .filter(db.KnowledgeBase.project_id.in_(list([p.id for p in projects])))
        )

        data = []
        project_names = {
            i.id: i.name
            for i in project_controller.get_list()
        }

        for record in query:
            vector_database = record.vector_database
            embedding_model = record.embedding_model

            data.append({
                'id': record.id,
                'name': record.name,
                'project_id': record.project_id,
                'project_name': project_names[record.project_id],
                'embedding_model': embedding_model.name if embedding_model is not None else None,
                'vector_database': None if vector_database is None else vector_database.name,
                'vector_database_table': record.vector_database_table,
                'params': record.params
            })

        return data

    def update(self, name: str, project_id: int, **kwargs) -> db.KnowledgeBase:
        """
        Update a knowledge base record
        """
        raise NotImplementedError()



from pydantic_settings import BaseSettings


class MindsDBInferenceHandlerConfig(BaseSettings):
    """
    Configuration for MindsDB Inference handler.

    Attributes
    ----------

    BASE_URL : str
        Base URL for the MindsDB Inference Endpoints API.
    """

    BASE_URL: str = "https://llm.mdb.ai/"


mindsdb_inference_handler_config = MindsDBInferenceHandlerConfig()



from collections import defaultdict

from flask import current_app as ca
from flask_restx import Resource

from mindsdb.api.http.utils import http_error
from mindsdb.api.http.namespaces.configs.tree import ns_conf
from mindsdb.metrics.metrics import api_endpoint_metrics


@ns_conf.route('/')
class GetRoot(Resource):
    @ns_conf.doc('get_tree_root')
    @api_endpoint_metrics('GET', '/tree')
    def get(self):
        databases = ca.database_controller.get_list()
        result = [{
            'name': x['name'],
            'class': 'db',
            'type': x['type'],
            'engine': x['engine'],
            'deletable': x['deletable'],
            'visible': x['visible']
        } for x in databases]
        return result


@ns_conf.route('/<db_name>')
@ns_conf.param('db_name', "Name of the database")
class GetLeaf(Resource):
    @ns_conf.doc('get_tree_leaf')
    @api_endpoint_metrics('GET', '/tree/database')
    def get(self, db_name):
        databases = ca.database_controller.get_dict()
        if db_name not in databases:
            return http_error(
                400,
                "Error",
                f"There is no element with name '{db_name}'"
            )
        db = databases[db_name]
        if db['type'] == 'project':
            project = ca.database_controller.get_project(db_name)
            tables = project.get_tables()
            tables = [{
                'name': key,
                'schema': None,
                'class': 'table',
                'type': val['type'],
                'engine': val.get('engine'),
                'deletable': val.get('deletable')
            } for key, val in tables.items()]
        elif db['type'] == 'data':
            handler = ca.integration_controller.get_data_handler(db_name)
            response = handler.get_tables()
            if response.type != 'table':
                return []
            table_types = {
                'BASE TABLE': 'table',
                'VIEW': 'view'
            }
            tables = response.data_frame.to_dict(orient='records')

            schemas = defaultdict(list)

            for table_meta in tables:
                table_meta = {key.lower(): val for key, val in table_meta.items()}
                schama = table_meta.get('table_schema')
                schemas[schama].append({
                    'name': table_meta['table_name'],
                    'class': 'table',
                    'type': table_types.get(table_meta.get('table_type')),
                    'engine': None,
                    'deletable': False
                })
            if len(schemas) == 1 and list(schemas.keys())[0] is None:
                tables = schemas[None]
            else:
                tables = [{
                    'name': key,
                    'class': 'schema',
                    'deletable': False,
                    'children': val
                } for key, val in schemas.items()]
        elif db['type'] == 'system':
            system_db = ca.database_controller.get_system_db(db_name)
            tables = system_db.get_tables()
            tables = [{
                'name': table.name,
                'class': table.kind,
                'type': 'system view',
                'engine': None,
                'deletable': table.deletable,
            } for table in tables.values()]
        return tables



from typing import Any

import sys


def format_exception_error(exception):
    try:
        exception_type, _exception_object, exception_traceback = sys.exc_info()
        filename = exception_traceback.tb_frame.f_code.co_filename
        line_number = exception_traceback.tb_lineno
        error_message = f'{exception_type.__name__}: {exception}, raised at: {filename}#{line_number}'
    except Exception:
        error_message = str(exception)
    return error_message


def dict_to_yaml(d, indent=0):
    yaml_str = ""
    for k, v in d.items():
        yaml_str += " " * indent + str(k) + ": "
        if isinstance(v, dict):
            yaml_str += "\n" + dict_to_yaml(v, indent + 2)
        else:
            yaml_str += str(v) + "\n"
    return yaml_str


# Mocks won't always have 'name' attribute.
def get_class_name(instance: Any, default: str = 'unknown'):
    if hasattr(instance.__class__, 'name'):
        return instance.__class__.name
    return default



from collections import OrderedDict

from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE


connection_args = OrderedDict(
    host={
        "type": ARG_TYPE.STR,
        "description": "The host name or IP address of the DB2 server/database.",
    },
    database={
        "type": ARG_TYPE.STR,
        "description": """
            The database name to use when connecting with the DB2 server.
        """,
    },
    user={
        "type": ARG_TYPE.STR,
        "description": "The user name used to authenticate with the DB2 server.",
    },
    password={
        "type": ARG_TYPE.PWD,
        "description": "The password to authenticate the user with the DB2 server.",
        'secret': True
    },
    port={
        "type": ARG_TYPE.INT,
        "description": "Specify port to connect DB2 through TCP/IP",
    },
    schemaName={"type": ARG_TYPE.STR, "description": "Specify the schema name "},
)

connection_args_example = OrderedDict(
    host="127.0.0.1",
    port="25000",
    password="1234",
    user="db2admin",
    schemaName="db2admin",
    database="BOOKS",
)



from mindsdb.integrations.handlers.instatus_handler.instatus_tables import StatusPages, Components
from mindsdb.integrations.libs.api_handler import APIHandler
from mindsdb.integrations.libs.response import HandlerStatusResponse as StatusResponse
from mindsdb.utilities import log
from mindsdb_sql import parse_sql
import requests
import pandas as pd
from collections import OrderedDict
from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE

logger = log.getLogger(__name__)


class InstatusHandler(APIHandler):
    def __init__(self, name: str, **kwargs) -> None:
        """initializer method

        Args:
            name (str): handler name
        """
        super().__init__(name)
        self._base_url = "https://api.instatus.com"
        self._api_key = None

        args = kwargs.get('connection_data', {})
        if 'api_key' in args:
            self._api_key = args['api_key']

        self.connection = None
        self.is_connected = False

        _tables = [
            StatusPages,
            Components
        ]

        for Table in _tables:
            self._register_table(Table.name, Table(self))

    def check_connection(self) -> StatusResponse:
        """checking the connection

        Returns:
            StatusResponse: whether the connection is still up
        """
        response = StatusResponse(False)

        try:
            self.connect()
            response.success = True
        except Exception as e:
            logger.error(f'Error connecting to Instatus API: {e}!')
            response.error_message = e

        self.is_connected = response.success
        return response

    def connect(self) -> StatusResponse:
        # If already connected, return the existing connection
        if self.is_connected and self.connection:
            return self.connection

        if self._api_key:
            try:
                headers = {"Authorization": f"Bearer {self._api_key}"}
                response = requests.get(f"{self._base_url}/v2/pages", headers=headers)

                if response.status_code == 200:
                    self.connection = response
                    self.is_connected = True
                    return StatusResponse(True)
                else:
                    raise Exception(f"Error connecting to Instatus API: {response.status_code} - {response.text}")
            except requests.RequestException as e:
                raise Exception(f"Request to Instatus API failed: {str(e)}")

        raise Exception("API key is missing")

    def native_query(self, query: str) -> StatusResponse:
        """Receive and process a raw query.

        Parameters
        ----------
        query : str
            query in a native format

        Returns
        -------
        StatusResponse
            Request status
        """
        ast = parse_sql(query, dialect="mindsdb")
        return self.query(ast)

    def call_instatus_api(self, endpoint: str, method: str = 'GET', params: dict = None, json_data: dict = {}) -> pd.DataFrame:
        if not params:
            params = {}

        headers = {"Authorization": f"Bearer {self._api_key}"}
        url = f"{self._base_url}{endpoint}"

        if method.upper() in ('GET', 'POST', 'PUT', 'DELETE'):
            headers['Content-Type'] = 'application/json'

            response = requests.request(method, url, headers=headers, params=params, json=json_data)

            if response.status_code == 200:
                data = response.json()
                return pd.DataFrame(data) if isinstance(data, list) else pd.DataFrame([data])
            else:
                raise Exception(f"Error connecting to Instatus API: {response.status_code} - {response.text}")

        return pd.DataFrame()


connection_args = OrderedDict(
    api_key={
        "type": ARG_TYPE.PWD,
        "description": "Instatus API key to use for authentication.",
        "required": True,
        "label": "Api key",
    },
)

connection_args_example = OrderedDict(
    api_key="d25509b171ad79395dc2c51b099ee6d0"
)



from mindsdb.integrations.libs.const import HANDLER_TYPE

from .__about__ import __description__ as description
from .__about__ import __version__ as version

try:
    from .lightwood_handler import LightwoodHandler as Handler

    import_error = None
except Exception as e:
    Handler = None
    import_error = e

title = 'Lightwood'
name = 'lightwood'
type = HANDLER_TYPE.ML
icon_path = "icon.svg"

__all__ = ['Handler', 'version', 'name', 'type', 'title', 'description', 'import_error', 'icon_path']



from collections import OrderedDict

from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE


connection_args = OrderedDict(
    server_hostname={
        "type": ARG_TYPE.STR,
        "description": "The server hostname for the cluster or SQL warehouse.",
        "required": True,
        "label": "Server_hostname",
    },
    http_path={
        "type": ARG_TYPE.STR,
        "description": "The HTTP path of the cluster or SQL warehouse.",
        "required": True,
        "label": "Http path",
    },
    access_token={
        "type": ARG_TYPE.STR,
        "description": "A Databricks personal access token for the workspace for the cluster or SQL warehouse.",
        "required": True,
        "label": "Access token",
        'secret': True
    },
    session_configuration={
        "type": ARG_TYPE.STR,
        "description": "A dictionary of Spark session configuration parameters. This parameter is optional.",
        "required": False,
        "label": "Session configuration",
    },
    http_headers={
        "type": ARG_TYPE.STR,
        "description": "Additional (key, value) pairs to set in HTTP headers on every RPC request the client makes."
        " This parameter is optional.",
        "required": False,
        "label": "Http headers",
    },
    catalog={
        "type": ARG_TYPE.STR,
        "description": "Catalog to use for the connection. This parameter is optional.",
        "required": False,
        "label": "Catalog",
    },
    schema={
        "type": ARG_TYPE.STR,
        "description": "Schema (database) to use for the connection. This parameter is optional.",
        "required": False,
        "label": "Schema",
    },
)

connection_args_example = OrderedDict(
    server_hostname="adb-1234567890123456.7.azuredatabricks.net",
    http_path="sql/protocolv1/o/1234567890123456/1234-567890-test123",
    access_token="dapi1234567890ab1cde2f3ab456c7d89efa",
    schema="sales",
)



from mindsdb.interfaces.storage import db
from mindsdb.utilities.context import context as ctx


def fmt_log_record(log_record):
    return {
        'log_from': 'mindsdb',
        'level': log_record.log_type,
        'context': 'unknown',
        'text': log_record.payload,
        'created_at': str(log_record.created_at).split('.')[0]
    }


def get_logs(min_timestamp, max_timestamp, context, level, log_from, limit):
    logs = db.session.query(db.Log).filter(
        db.Log.company_id == ctx.company_id,
        db.Log.created_at > min_timestamp
    )

    if max_timestamp is not None:
        logs = logs.filter(db.Log.created_at < max_timestamp)

    if context is not None:
        # e.g. datasource/predictor and assoicated id
        pass

    if level is not None:
        logs = logs.filter(db.Log.log_type == level)

    if log_from is not None:
        # mindsdb/native/lightwood/all
        pass

    if limit is not None:
        logs = logs.limit(limit)

    logs = [fmt_log_record(x) for x in logs]
    return logs



from mindsdb.integrations.handlers.youtube_handler.youtube_tables import (
    YoutubeCommentsTable,
    YoutubeChannelsTable,
    YoutubeVideosTable,
)
from mindsdb.integrations.libs.api_handler import APIHandler
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
)
from mindsdb.utilities import log
from mindsdb_sql import parse_sql

from collections import OrderedDict
from mindsdb.utilities.config import Config
from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE

from googleapiclient.discovery import build

from mindsdb.integrations.utilities.handlers.auth_utilities import GoogleUserOAuth2Manager

DEFAULT_SCOPES = [
	'https://www.googleapis.com/auth/youtube',
	'https://www.googleapis.com/auth/youtube.force-ssl',
	'https://www.googleapis.com/auth/youtubepartner'
]

logger = log.getLogger(__name__)


class YoutubeHandler(APIHandler):
    """Youtube handler implementation"""

    def __init__(self, name=None, **kwargs):
        """Initialize the Youtube handler.
        Parameters
        ----------
        name : str
            name of a handler instance
        """
        super().__init__(name)

        connection_data = kwargs.get("connection_data", {})

        self.parser = parse_sql
        self.connection_data = connection_data
        self.kwargs = kwargs
        self.connection = None
        self.is_connected = False

        self.handler_storage = kwargs['handler_storage']

        self.credentials_url = self.connection_data.get('credentials_url', None)
        self.credentials_file = self.connection_data.get('credentials_file', None)
        if self.connection_data.get('credentials'):
            self.credentials_file = self.connection_data.pop('credentials')
        if not self.credentials_file and not self.credentials_url:
            # try to get from config
            yt_config = Config().get('handlers', {}).get('youtube', {})
            secret_file = yt_config.get('credentials_file')
            secret_url = yt_config.get('credentials_url')
            if secret_file:
                self.credentials_file = secret_file
            elif secret_url:
                self.credentials_url = secret_url

        self.youtube_api_token = self.connection_data.get('youtube_api_token', None)

        self.scopes = self.connection_data.get('scopes', DEFAULT_SCOPES)

        youtube_video_comments_data = YoutubeCommentsTable(self)
        self._register_table("comments", youtube_video_comments_data)

        youtube_channel_data = YoutubeChannelsTable(self)
        self._register_table("channels", youtube_channel_data)

        youtube_video_data = YoutubeVideosTable(self)
        self._register_table("videos", youtube_video_data)

    def connect(self) -> StatusResponse:
        """Set up the connection required by the handler.
        Returns
        -------
        StatusResponse
            connection object
        """
        if self.is_connected is True:
            return self.connection
        
        google_oauth2_manager = GoogleUserOAuth2Manager(self.handler_storage, self.scopes, self.credentials_file, self.credentials_url, self.connection_data.get('code'))
        creds = google_oauth2_manager.get_oauth2_credentials()

        youtube = build(
            "youtube", "v3", developerKey=self.youtube_api_token, credentials=creds
        )
        self.connection = youtube

        return self.connection

    def check_connection(self) -> StatusResponse:
        """Check connection to the handler.
        Returns
        -------
        StatusResponse
            Status confirmation
        """
        response = StatusResponse(False)
        need_to_close = self.is_connected is False

        try:
            self.connect()
            response.success = True
            response.copy_storage = True
        except Exception as e:
            logger.error(f"Error connecting to Youtube API: {e}!")
            response.error_message = e

        self.is_connected = response.success

        return response

    def native_query(self, query: str) -> StatusResponse:
        """Receive and process a raw query.
        Parameters
        ----------
        query : str
            query in a native format
        Returns
        -------
        StatusResponse
            Request status
        """
        ast = parse_sql(query, dialect="mindsdb")
        return self.query(ast)


connection_args = OrderedDict(
    youtube_access_token={
        "type": ARG_TYPE.STR,
        "description": "API Key",
        "label": "API Key",
    },
    credentials_url={
        'type': ARG_TYPE.STR,
        'description': 'URL to OAuth2 Credentials',
        'label': 'URL to OAuth2 Credentials',
    },
    credentials_file={
        'type': ARG_TYPE.STR,
        'description': 'Location of OAuth2 Credentials',
        'label': 'Location of OAuth2 Credentials',
    },
    credentials={
        'type': ARG_TYPE.PATH,
        'description': 'OAuth2 Credentials',
        'label': 'Upload OAuth2 Credentials',
    },
    code={
        'type': ARG_TYPE.STR,
        'description': 'Authentication Code',
        'label': 'Authentication Code',
    }
)

connection_args_example = OrderedDict(youtube_api_token="<your-youtube-api-token>")



"""
    Tasks queue allows to limit load by ordering tasks through a queue.
    Current implementation use Redis as backend for the queue. To run MindsDB with tasks queue need:
    1. config mindsdb to use tasks queue by one of:
        - fill 'ml_task_queue' key in config.json:
            {
                "ml_task_queue": {
                    "type": "redis",    // required
                    "host": "...",
                    "port": "...",
                    "db": "...",
                    "username": "...",
                    "password": "..."
                }
            }
        - or set env vars:
            MINDSDB_ML_QUEUE_TYPE=redis  # required
            MINDSDB_ML_QUEUE_HOST=...
            MINDSDB_ML_QUEUE_PORT=...
            MINDSDB_ML_QUEUE_DB=...
            MINDSDB_ML_QUEUE_USERNAME=...
            MINDSDB_ML_QUEUE_PASSWORD=...
    2. run mindsdb with arg --ml_task_queue_consumer

    In redis there is two types of entities used: streams (for distributing tasks) and regular
    key-value storage with ttl (to transfer dataframes and some other data). Dataframes are not
    transfer via streams to make stream messages lightweight.

    Taks queue may work in single instnace to limit load on it, ot it may work in distributed
    system. In that case mindsdb may be splitted into two modules: parser/planner/executioner (PPE)
    and ML.

  ┌─────────────┐ ┌─────────────┐
  │             │ │             │
  │ MindsDB PPE │ │ MindsDB PPE │
  │             │ │             │
  └───────────┬─┘ └─┬─────┬─▲───┘
              │     │     │ │
             ┌▼─────▼┐    │ │
             │       │  ┌─▼─┴─────────┐
             │ Queue │  │             │
             │       │  │  Cache      │
             ├───────┤  │             │
             │ Task  │  ├─────────────┤
             ├───────┤  │  Dataframe  │
             │ Task  │  ├─────────────┤
             ├───────┤  │  Status     │
             │ Task  │  └─┬─▲─────────┘
             └┬─────┬┘    │ │
              │     │     │ │
              │     │     │ │
  ┌───────────▼─┐ ┌─▼─────▼─┴───┐
  │             │ │             │
  │ MindsDB ML  │ │ MindsDB ML  │
  │             │ │             │
  └─────────────┘ └─────────────┘
"""


from mindsdb.utilities import log

from mindsdb.integrations.libs.api_handler import APIHandler

from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
)
from mindsdb_sql import parse_sql

from mindsdb.integrations.handlers.email_handler.email_tables import EmailsTable
from mindsdb.integrations.handlers.email_handler.email_client import EmailClient
from mindsdb.integrations.handlers.email_handler.settings import EmailConnectionDetails


logger = log.getLogger(__name__)


class EmailHandler(APIHandler):
    """
    A class for handling connections and interactions with Email (send and search).

    Parameters
    ----------
    name : str
        The name of the handler
    connection_data : EmailConnectionDetails
        The connection details for the email server

        see `EmailConnectionDetails` for more details and examples

    """

    def __init__(self, name=None, **kwargs):
        super().__init__(name)

        connection_data = kwargs.get("connection_data", {})
        self.connection_data = EmailConnectionDetails(**connection_data)
        self.kwargs = kwargs

        self.connection = None
        self.is_connected = False

        emails = EmailsTable(self)
        self._register_table('emails', emails)

    def connect(self):
        """Authenticate with the email servers using credentials."""

        if self.is_connected is True:
            return self.connection

        try:
            self.connection = EmailClient(self.connection_data)
        except Exception as e:
            logger.error(f'Error connecting to email api: {e}!')
            raise e

        self.is_connected = True
        return self.connection

    def disconnect(self):
        """ Close any existing connections

        Should switch self.is_connected.
        """
        self.is_connected = False

        return self.connection.logout()

    def check_connection(self) -> StatusResponse:

        response = StatusResponse(False)

        try:
            self.connect()
            response.success = True

        except Exception as e:
            response.error_message = f'Error connecting to Email: {e}. '
            logger.error(response.error_message)

        if response.success is False and self.is_connected is True:
            self.is_connected = False

        return response

    def native_query(self, query: str) -> StatusResponse:
        """Receive and process a raw query.
        Parameters
        ----------
        query : str
            query in a native format
        Returns
        -------
        StatusResponse
            Request status
        """
        ast = parse_sql(query, dialect="mindsdb")
        return self.query(ast)



import re
import time

from bson import ObjectId
import certifi
import pandas as pd
import pymongo
from pymongo import MongoClient

from mindsdb_sql.parser.ast.base import ASTNode

from mindsdb.utilities import log
from mindsdb.integrations.libs.base import DatabaseHandler
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
    RESPONSE_TYPE
)
from .utils.mongodb_render import MongodbRender
from mindsdb.api.mongo.utilities.mongodb_query import MongoQuery
from mindsdb.api.mongo.utilities.mongodb_parser import MongodbParser


logger = log.getLogger(__name__)

class MongoDBHandler(DatabaseHandler):
    """
    This handler handles connection and execution of the MongoDB statements.
    """
    _SUBSCRIBE_SLEEP_INTERVAL = 0.5

    name = 'mongodb'

    def __init__(self, name, **kwargs):
        super().__init__(name)
        connection_data = kwargs['connection_data']
        self.host = connection_data.get("host")
        self.port = int(connection_data.get("port") or 27017)
        self.user = connection_data.get("username")
        self.password = connection_data.get("password")
        self.database = connection_data.get('database')
        self.flatten_level = connection_data.get('flatten_level', 0)

        self.connection = None
        self.is_connected = False

    def __del__(self):
        if self.is_connected is True:
            self.disconnect()

    def connect(self):
        kwargs = {}
        if isinstance(self.user, str) and len(self.user) > 0:
            kwargs['username'] = self.user

        if isinstance(self.password, str) and len(self.password) > 0:
            kwargs['password'] = self.password

        if re.match(r'/?.*tls=true', self.host.lower()):
            kwargs['tls'] = True

        if re.match(r'/?.*tls=false', self.host.lower()):
            kwargs['tls'] = False

        if re.match(r'.*.mongodb.net', self.host.lower()) and kwargs.get('tls', None) is None:
            kwargs['tlsCAFile'] = certifi.where()
            if kwargs.get('tls', None) is None:
                kwargs['tls'] = True

        connection = MongoClient(
            self.host,
            port=self.port,
            **kwargs
        )

        # detect database from connection
        if self.database is None:
            self.database = connection.get_database().name

        self.is_connected = True
        self.connection = connection
        return self.connection

    def subscribe(self, stop_event, callback, table_name, columns=None, **kwargs):

        con = self.connect()
        cur = con[self.database][table_name].watch()
        while True:
            if stop_event.is_set():
                cur.close()
                return

            res = cur.try_next()
            if res is None:
                time.sleep(self._SUBSCRIBE_SLEEP_INTERVAL)
                continue
            _id = res['documentKey']['_id']
            if res['operationType'] == 'insert':
                if columns is not None:
                    updated_columns = set(res['fullDocument'].keys())
                    if not set(columns) & set(updated_columns):
                        # do nothing
                        continue

                callback(row=res['fullDocument'], key={'_id': _id})
            if res['operationType'] == 'update':
                if columns is not None:
                    updated_columns = set(res['updateDescription']['updatedFields'].keys())
                    if not set(columns) & set(updated_columns):
                        # do nothing
                        continue

                # get all document
                full_doc = con[self.database][table_name].find_one(res['documentKey'])
                callback(row=full_doc, key={'_id': _id})

    def disconnect(self):
        if self.is_connected:
            self.connection.close()
            self.is_connected = False


    def check_connection(self) -> StatusResponse:
        """
        Check the connection of the database
        :return: success status and error message if error occurs
        """

        result = StatusResponse(False)
        need_to_close = self.is_connected is False

        try:
            con = self.connect()
            con.server_info()
            result.success = True
        except Exception as e:
            logger.error(f'Error connecting to MongoDB {self.database}, {e}!')
            result.error_message = str(e)

        if result.success is True and need_to_close:
            self.disconnect()
        if result.success is False and self.is_connected is True:
            self.is_connected = False

        return result

    def native_query(self, query: [str, MongoQuery, dict]) -> Response:

        """
        input str or MongoQuery

        returns the records from the current recordset
        """
        if isinstance(query, str):
            query = MongodbParser().from_string(query)

        if isinstance(query, dict):
            # failback for previous api

            mquery = MongoQuery(query['collection'])

            for c in  query['call']:
                mquery.add_step({
                    'method': c['method'],
                    'args': c['args']
                })

            query = mquery

        collection = query.collection
        database = self.database

        con = self.connect()

        try:

            cursor = con[database][collection]

            for step in query.pipeline:
                fnc = getattr(cursor, step['method'])
                cursor = fnc(*step['args'])

            result = []
            if not isinstance(cursor, pymongo.results.UpdateResult):
                for row in cursor:
                    result.append(self.flatten(row, level=self.flatten_level))

            if len(result) > 0:
                df = pd.DataFrame(result)
            else:
                columns = list(self.get_columns(collection).data_frame.Field)
                df = pd.DataFrame([], columns=columns)

            response = Response(
                RESPONSE_TYPE.TABLE,
                df
            )

        except Exception as e:
            logger.error(f'Error running query: {query} on {self.database}.{collection}!')
            response = Response(
                RESPONSE_TYPE.ERROR,
                error_message=str(e)
            )

        return response

    def flatten(self, row, level=0):
        # move sub-keys to upper level

        add = {}
        del_keys = []
        edit_keys = {}
        for k, v in row.items():
            # convert objectId to string
            if isinstance(v, ObjectId):
                edit_keys[k] = str(v)
            if level > 0:
                if isinstance(v, dict):
                    for k2, v2 in self.flatten(v, level=level - 1).items():
                        add[f'{k}.{k2}'] = v2
                    del_keys.append(k)
        if add:
            row.update(add)
        for key in del_keys:
            del row[key]
        if edit_keys:
            row.update(edit_keys)

        return row

    def query(self, query: ASTNode) -> Response:
        """
        Retrieve the data from the SQL statement.
        """
        renderer = MongodbRender()
        mquery = renderer.to_mongo_query(query)
        return self.native_query(mquery)

    def get_tables(self) -> Response:
        """
        Get a list with of collection in database
        """
        con = self.connect()
        collections = con[self.database].list_collection_names()
        collections_ar = [
            [i] for i in collections
        ]
        df = pd.DataFrame(collections_ar, columns=['table_name'])

        response = Response(
            RESPONSE_TYPE.TABLE,
            df
        )

        return response

    def get_columns(self, collection) -> Response:
        """
        Use first row to detect columns
        """
        con = self.connect()
        record = con[self.database][collection].find_one()

        data = []
        if record is not None:
            record = self.flatten(record)

            for k, v in record.items():
                data.append([k, type(v).__name__])

        df = pd.DataFrame(data, columns=['Field', 'Type'])

        response = Response(
            RESPONSE_TYPE.TABLE,
            df
        )
        return response



from influxdb_client_3 import InfluxDBClient3
from mindsdb.integrations.handlers.influxdb_handler.influxdb_tables import InfluxDBTables
from mindsdb.integrations.libs.api_handler import APIHandler
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
)
from mindsdb.utilities import log
from mindsdb_sql import parse_sql



logger = log.getLogger(__name__)

class InfluxDBHandler(APIHandler):
    """InfluxDB handler implementation"""

    def __init__(self, name=None, **kwargs):
        """Initialize the InfluxDB handler.
        Parameters
        ----------
        name : str
            name of a handler instance
        """
        super().__init__(name)

        connection_data = kwargs.get("connection_data", {})

        self.parser = parse_sql
        self.dialect = 'influxdb'
        self.connection_data = connection_data
        self.kwargs = kwargs
        self.connection = None
        self.is_connected = False

        influxdb_tables_data = InfluxDBTables(self)
        self._register_table("tables", influxdb_tables_data)
        

    def connect(self):
        """Set up the connection required by the handler.
        Returns
        -------
        None

        Raises Expection if ping check fails
        """
        
        if self.is_connected is True:
            return self.connection
        
        self.connection=InfluxDBClient3(host=self.connection_data['influxdb_url'],token=self.connection_data['influxdb_token'],org=self.connection_data.get('org'))


        self.is_connected = True

        return self.connection
        

    def check_connection(self) -> StatusResponse:
        """Check connection to the handler.
        Returns
        -------
        StatusResponse
            Status confirmation
        """
        response = StatusResponse(False)

        try:
            self.connect()
            response.success = True
        except Exception as e:
            logger.error(f"Error connecting to InfluxDB API: {e}!")
            response.error_message = e

        self.is_connected = response.success

        return response

    def native_query(self, query: str) -> StatusResponse:
        """Receive and process a raw query.
        Parameters
        ----------
        query : str
            query in a native format
        Returns
        -------
        StatusResponse
            Request status
        """
        ast = parse_sql(query, dialect="mindsdb")
        return self.query(ast)

    def call_influxdb_tables(self,query):
        """Pulls all the records from the given  InfluxDB table and returns it select()
    
        Returns
        -------
        pd.DataFrame of all the records of the particular InfluxDB 
        """
        influx_connection=self.connect()
        if query is None:
            query='SELECT * FROM '+ f"{self.connection_data['influxdb_table_name']}"
        
        table=influx_connection.query(query=query,database=self.connection_data['influxdb_db_name'], language='sql')
        return table.to_pandas()



from ckanapi import RemoteCKAN as rc
import pandas as pd

from mindsdb.integrations.libs.base import DatabaseHandler
from mindsdb.integrations.libs.response import HandlerStatusResponse, HandlerResponse, RESPONSE_TYPE
from mindsdb_sql.parser.ast.base import ASTNode
from mindsdb_sql.render.sqlalchemy_render import SqlalchemyRender
from mindsdb.utilities import log

logger = log.getLogger(__name__)


class CkanHandler(DatabaseHandler):
    """
    This handler handles connection and consumption of the CKAN API.
    """
    name = "ckan"

    def __init__(self, name=None, **kwargs):
        super().__init__(name)
        self.type = 'ckan'
        self.name = name
        self.connection_args = kwargs.get('connection_data')
        self.ckan = None
        self.dialect = 'postgresql'
        self.renderer = SqlalchemyRender('postgres')
        self.connection = None
        self.is_connected = False

    def connect(self) -> HandlerStatusResponse:
        """
        Handles the connection to a CKAN remote portal instance.
        """
        url = self.connection_args.get('url')
        try:
            ckan = rc(url)
            self.is_connected = True
            self.ckan = ckan
        except Exception as e:
            return HandlerStatusResponse(False, f'Failed to connect to CKAN: {e}')
        self.connection = ckan
        return HandlerStatusResponse(True)

    def disconnect(self):
        self.is_connected = False

    def check_connection(self) -> HandlerStatusResponse:
        response = HandlerStatusResponse(False)
        try:
            self.connect()
            result = self.ckan.action.status_show()
            if 'datastore' not in result.get('extensions'):
                """
                If the CKAN instance does not have the datastore extension,
                we can't use it.
                """
                response.message = 'CKAN datastore is not enabled'
                response.status = False
                self.is_connected = False
                return response

        except Exception as e:
            logger.error(f'Error connecting to CKAN: {e}!')
            self.is_connected = False
            response.error_message = e

        if response.success is False and self.is_connected is True:
            self.is_connected = False

        response.success = True
        return response

    def query(self, query: ASTNode) -> HandlerResponse:
        if not self.ckan:
            self.connect()
        query_str = self.renderer.get_string(query, with_failback=True)
        return self.native_query(query_str)

    def native_query(self, query_str: str) -> HandlerResponse:
        """
        Execute a query on the CKAN instance.
        :param query_str: The query to execute.
        :return: The response of the query.
        """
        if not self.ckan:
            self.connect()
        result = self.ckan.action.datastore_search_sql(sql=query_str)
        if len(result.get('records')) > 0:
            df = pd.DataFrame(result['records'])
            response = HandlerResponse(RESPONSE_TYPE.TABLE, df)
        else:
            response = HandlerResponse(RESPONSE_TYPE.TABLE, None)
        return response

    def get_tables(self) -> HandlerResponse:
        if not self.ckan:
            self.connect()
        result = self.ckan.action.datastore_search(resource_id='_table_metadata')
        df = pd.DataFrame(result['records'])
        response = HandlerResponse(RESPONSE_TYPE.TABLE, df)
        return response
